# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Some common functions and types.

from binascii import hexlify as _hexlify
from collections import namedtuple

import yaml as _yaml

is_string = lambda d: isinstance(d, str)

#
#
#


def crc16(data: bytes):
    '''
    CRC-16 (CCITT) implemented with a precomputed lookup table
    '''
    table = [
        0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50A5, 0x60C6, 0x70E7, 0x8108, 0x9129, 0xA14A, 0xB16B, 0xC18C, 0xD1AD, 0xE1CE,
        0xF1EF, 0x1231, 0x0210, 0x3273, 0x2252, 0x52B5, 0x4294, 0x72F7, 0x62D6, 0x9339, 0x8318, 0xB37B, 0xA35A, 0xD3BD, 0xC39C,
        0xF3FF, 0xE3DE, 0x2462, 0x3443, 0x0420, 0x1401, 0x64E6, 0x74C7, 0x44A4, 0x5485, 0xA56A, 0xB54B, 0x8528, 0x9509, 0xE5EE,
        0xF5CF, 0xC5AC, 0xD58D, 0x3653, 0x2672, 0x1611, 0x0630, 0x76D7, 0x66F6, 0x5695, 0x46B4, 0xB75B, 0xA77A, 0x9719, 0x8738,
        0xF7DF, 0xE7FE, 0xD79D, 0xC7BC, 0x48C4, 0x58E5, 0x6886, 0x78A7, 0x0840, 0x1861, 0x2802, 0x3823, 0xC9CC, 0xD9ED, 0xE98E,
        0xF9AF, 0x8948, 0x9969, 0xA90A, 0xB92B, 0x5AF5, 0x4AD4, 0x7AB7, 0x6A96, 0x1A71, 0x0A50, 0x3A33, 0x2A12, 0xDBFD, 0xCBDC,
        0xFBBF, 0xEB9E, 0x9B79, 0x8B58, 0xBB3B, 0xAB1A, 0x6CA6, 0x7C87, 0x4CE4, 0x5CC5, 0x2C22, 0x3C03, 0x0C60, 0x1C41, 0xEDAE,
        0xFD8F, 0xCDEC, 0xDDCD, 0xAD2A, 0xBD0B, 0x8D68, 0x9D49, 0x7E97, 0x6EB6, 0x5ED5, 0x4EF4, 0x3E13, 0x2E32, 0x1E51, 0x0E70,
        0xFF9F, 0xEFBE, 0xDFDD, 0xCFFC, 0xBF1B, 0xAF3A, 0x9F59, 0x8F78, 0x9188, 0x81A9, 0xB1CA, 0xA1EB, 0xD10C, 0xC12D, 0xF14E,
        0xE16F, 0x1080, 0x00A1, 0x30C2, 0x20E3, 0x5004, 0x4025, 0x7046, 0x6067, 0x83B9, 0x9398, 0xA3FB, 0xB3DA, 0xC33D, 0xD31C,
        0xE37F, 0xF35E, 0x02B1, 0x1290, 0x22F3, 0x32D2, 0x4235, 0x5214, 0x6277, 0x7256, 0xB5EA, 0xA5CB, 0x95A8, 0x8589, 0xF56E,
        0xE54F, 0xD52C, 0xC50D, 0x34E2, 0x24C3, 0x14A0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405, 0xA7DB, 0xB7FA, 0x8799, 0x97B8,
        0xE75F, 0xF77E, 0xC71D, 0xD73C, 0x26D3, 0x36F2, 0x0691, 0x16B0, 0x6657, 0x7676, 0x4615, 0x5634, 0xD94C, 0xC96D, 0xF90E,
        0xE92F, 0x99C8, 0x89E9, 0xB98A, 0xA9AB, 0x5844, 0x4865, 0x7806, 0x6827, 0x18C0, 0x08E1, 0x3882, 0x28A3, 0xCB7D, 0xDB5C,
        0xEB3F, 0xFB1E, 0x8BF9, 0x9BD8, 0xABBB, 0xBB9A, 0x4A75, 0x5A54, 0x6A37, 0x7A16, 0x0AF1, 0x1AD0, 0x2AB3, 0x3A92, 0xFD2E,
        0xED0F, 0xDD6C, 0xCD4D, 0xBDAA, 0xAD8B, 0x9DE8, 0x8DC9, 0x7C26, 0x6C07, 0x5C64, 0x4C45, 0x3CA2, 0x2C83, 0x1CE0, 0x0CC1,
        0xEF1F, 0xFF3E, 0xCF5D, 0xDF7C, 0xAF9B, 0xBFBA, 0x8FD9, 0x9FF8, 0x6E17, 0x7E36, 0x4E55, 0x5E74, 0x2E93, 0x3EB2, 0x0ED1,
        0x1EF0
    ]

    crc = 0xFFFF
    for byte in data:
        crc = (crc << 8) ^ table[(crc >> 8) ^ byte]
        crc &= 0xFFFF  # important, crc must stay 16bits all the way through
    return crc


class NamedInt(int):
    """A regular Python integer with an attached name.

    Caution: comparison with strings will also match this NamedInt's name
    (case-insensitive)."""

    def __new__(cls, value, name):
        assert is_string(name)
        obj = int.__new__(cls, value)
        obj.name = str(name)
        return obj

    def bytes(self, count=2):
        return int2bytes(self, count)

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, NamedInt):
            return int(self) == int(other) and self.name == other.name
        if isinstance(other, int):
            return int(self) == int(other)
        if is_string(other):
            return self.name.lower() == other.lower()
        # this should catch comparisons with bytes in Py3
        if other is not None:
            raise TypeError('Unsupported type ' + str(type(other)))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return int(self)

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'NamedInt(%d, %r)' % (int(self), self.name)

    @classmethod
    def from_yaml(cls, loader, node):
        args = loader.construct_mapping(node)
        return cls(value=args['value'], name=args['name'])

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_mapping('!NamedInt', {'value': int(data), 'name': data.name}, flow_style=True)


_yaml.SafeLoader.add_constructor('!NamedInt', NamedInt.from_yaml)
_yaml.add_representer(NamedInt, NamedInt.to_yaml)


class NamedInts:
    """An ordered set of NamedInt values.

    Indexing can be made by int or string, and will return the corresponding
    NamedInt if it exists in this set, or `None`.

    Extracting slices will return all present NamedInts in the given interval
    (extended slices are not supported).

    Assigning a string to an indexed int will create a new NamedInt in this set;
    if the value already exists in the set (int or string), ValueError will be
    raised.
    """
    __slots__ = ('__dict__', '_values', '_indexed', '_fallback', '_is_sorted')

    def __init__(self, dict=None, **kwargs):

        def _readable_name(n):
            if not is_string(n):
                raise TypeError('expected string, got ' + str(type(n)))
            return n.replace('__', '/').replace('_', ' ')

        # print (repr(kwargs))
        elements = dict if dict else kwargs
        values = {k: NamedInt(v, _readable_name(k)) for (k, v) in elements.items()}
        self.__dict__ = values
        self._is_sorted = False
        self._values = list(values.values())
        self._sort_values()
        self._indexed = {int(v): v for v in self._values}
        # assert len(values) == len(self._indexed)
        # "(%d) %r\n=> (%d) %r" % (len(values), values, len(self._indexed), self._indexed)
        self._fallback = None

    @classmethod
    def list(cls, items, name_generator=lambda x: str(x)):  # noqa: B008
        values = {name_generator(x): x for x in items}
        return NamedInts(**values)

    @classmethod
    def range(cls, from_value, to_value, name_generator=lambda x: str(x), step=1):  # noqa: B008
        values = {name_generator(x): x for x in range(from_value, to_value + 1, step)}
        return NamedInts(**values)

    def flag_names(self, value):
        unknown_bits = value
        for k in self._indexed:
            assert bin(k).count('1') == 1
            if k & value == k:
                unknown_bits &= ~k
                yield str(self._indexed[k])

        if unknown_bits:
            yield 'unknown:%06X' % unknown_bits

    def _sort_values(self):
        self._values = sorted(self._values)
        self._is_sorted = True

    def __getitem__(self, index):
        if isinstance(index, int):
            if index in self._indexed:
                return self._indexed[int(index)]
            if self._fallback:
                value = NamedInt(index, self._fallback(index))
                self._indexed[index] = value
                self._values.append(value)
                self._is_sorted = False
                self._sort_values()
                return value

        elif is_string(index):
            if index in self.__dict__:
                return self.__dict__[index]
            return (next((x for x in self._values if str(x) == index), None))

        elif isinstance(index, slice):
            values = self._values if self._is_sorted else sorted(self._values)

            if index.start is None and index.stop is None:
                return values[:]

            v_start = int(values[0]) if index.start is None else int(index.start)
            v_stop = (values[-1] + 1) if index.stop is None else int(index.stop)

            if v_start > v_stop or v_start > values[-1] or v_stop <= values[0]:
                return []

            if v_start <= values[0] and v_stop > values[-1]:
                return values[:]

            start_index = 0
            stop_index = len(values)

            for i, value in enumerate(values):
                if value < v_start:
                    start_index = i + 1
                elif index.stop is None:
                    break
                if value >= v_stop:
                    stop_index = i
                    break

            return values[start_index:stop_index]

    def __setitem__(self, index, name):
        assert isinstance(index, int), type(index)
        if isinstance(name, NamedInt):
            assert int(index) == int(name), repr(index) + ' ' + repr(name)
            value = name
        elif is_string(name):
            value = NamedInt(index, name)
        else:
            raise TypeError('name must be a string')

        if str(value) in self.__dict__:
            raise ValueError('%s (%d) already known' % (value, int(value)))
        if int(value) in self._indexed:
            raise ValueError('%d (%s) already known' % (int(value), value))

        self._values.append(value)
        self._is_sorted = False
        self._sort_values()
        self.__dict__[str(value)] = value
        self._indexed[int(value)] = value

    def __contains__(self, value):
        if isinstance(value, NamedInt):
            return self[value] == value
        elif isinstance(value, int):
            return value in self._indexed
        elif is_string(value):
            return value in self.__dict__ or value in self._values

    def __iter__(self):
        yield from self._values

    def __len__(self):
        return len(self._values)

    def __repr__(self):
        return 'NamedInts(%s)' % ', '.join(repr(v) for v in self._values)

    def __or__(self, other):
        return NamedInts(**self.__dict__, **other.__dict__)


class UnsortedNamedInts(NamedInts):

    def _sort_values(self):
        pass

    def __or__(self, other):
        c = UnsortedNamedInts if isinstance(other, UnsortedNamedInts) else NamedInts
        return c(**self.__dict__, **other.__dict__)


def strhex(x):
    assert x is not None
    """Produce a hex-string representation of a sequence of bytes."""
    return _hexlify(x).decode('ascii').upper()


def bytes2int(x, signed=False):
    return int.from_bytes(x, signed=signed, byteorder='big')


def int2bytes(x, count=None, signed=False):
    if count:
        return x.to_bytes(length=count, byteorder='big', signed=signed)
    else:
        return x.to_bytes(length=8, byteorder='big', signed=signed).lstrip(b'\x00')


class KwException(Exception):
    """An exception that remembers all arguments passed to the constructor.
    They can be later accessed by simple member access.
    """

    def __init__(self, **kwargs):
        super().__init__(kwargs)

    def __getattr__(self, k):
        try:
            return super().__getattr__(k)
        except AttributeError:
            return self.args[0].get(k)  # was self.args[0][k]


"""Firmware information."""
FirmwareInfo = namedtuple('FirmwareInfo', ['kind', 'name', 'version', 'extras'])

BATTERY_APPROX = NamedInts(empty=0, critical=5, low=20, good=50, full=90)

del namedtuple



# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import json as _json
import logging
import os as _os
import os.path as _path

from threading import Lock as _Lock
from threading import Timer as _Timer

import yaml as _yaml

from logitech_receiver.common import NamedInt as _NamedInt
from solaar import __version__

logger = logging.getLogger(__name__)

_XDG_CONFIG_HOME = _os.environ.get('XDG_CONFIG_HOME') or _path.expanduser(_path.join('~', '.config'))
_yaml_file_path = _path.join(_XDG_CONFIG_HOME, 'solaar', 'config.yaml')
_json_file_path = _path.join(_XDG_CONFIG_HOME, 'solaar', 'config.json')

_KEY_VERSION = '_version'
_KEY_NAME = '_NAME'
_KEY_WPID = '_wpid'
_KEY_SERIAL = '_serial'
_KEY_MODEL_ID = '_modelId'
_KEY_UNIT_ID = '_unitId'
_KEY_ABSENT = '_absent'
_KEY_SENSITIVE = '_sensitive'
_config = []


def _load():
    loaded_config = []
    if _path.isfile(_yaml_file_path):
        path = _yaml_file_path
        try:
            with open(_yaml_file_path) as config_file:
                loaded_config = _yaml.safe_load(config_file)
        except Exception as e:
            logger.error('failed to load from %s: %s', _yaml_file_path, e)
    elif _path.isfile(_json_file_path):
        path = _json_file_path
        try:
            with open(_json_file_path) as config_file:
                loaded_config = _json.load(config_file)
        except Exception as e:
            logger.error('failed to load from %s: %s', _json_file_path, e)
        loaded_config = _convert_json(loaded_config)
    else:
        path = None
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('load => %s', loaded_config)
    global _config
    _config = _parse_config(loaded_config, path)


def _parse_config(loaded_config, config_path):
    current_version = __version__
    parsed_config = [current_version]
    try:
        if not loaded_config:
            return parsed_config

        loaded_version = loaded_config[0]
        discard_derived_properties = loaded_version != current_version
        if discard_derived_properties:
            if logger.isEnabledFor(logging.INFO):
                logger.info(
                    'config file \'%s\' was generated by another version of solaar '
                    '(config: %s, current: %s). refreshing detected device capabilities', config_path, loaded_version,
                    current_version
                )

        for device in loaded_config[1:]:
            assert isinstance(device, dict)
            parsed_config.append(_device_entry_from_config_dict(device, discard_derived_properties))
    except Exception as e:
        logger.warning('Exception processing config file \'%s\', ignoring contents: %s', config_path, e)
    return parsed_config


def _device_entry_from_config_dict(data, discard_derived_properties):
    divert = data.get('divert-keys')
    if divert:
        sliding = data.get('dpi-sliding')
        if sliding:  # convert old-style dpi-sliding setting to divert-keys entry
            divert[int(sliding)] = 3
        data.pop('dpi-sliding', None)
        gestures = data.get('mouse-gestures')
        if gestures:  # convert old-style mouse-gestures setting to divert-keys entry
            divert[int(gestures)] = 2
        data.pop('mouse-gestures', None)
        # remove any string entries (from bad conversions)
        data['divert-keys'] = {k: v for k, v in divert.items() if isinstance(k, int)}
    if data.get('_sensitive', None) is None:  # make scroll wheel settings default to ignore
        data['_sensitive'] = {
            'hires-smooth-resolution': 'ignore',
            'hires-smooth-invert': 'ignore',
            'hires-scroll-mode': 'ignore'
        }
    if discard_derived_properties:
        data.pop('_absent', None)
        data.pop('_battery', None)
    return _DeviceEntry(**data)


save_timer = None
save_lock = _Lock()
defer_saves = False  # don't allow configuration saves to be deferred


def save(defer=False):
    global save_timer
    if not _config:
        return
    dirname = _os.path.dirname(_yaml_file_path)
    if not _path.isdir(dirname):
        try:
            _os.makedirs(dirname)
        except Exception:
            logger.error('failed to create %s', dirname)
            return
    if not defer or not defer_saves:
        do_save()
    else:
        with save_lock:
            if not save_timer:
                from gi.repository import GLib
                save_timer = _Timer(5.0, lambda: GLib.idle_add(do_save))
                save_timer.start()


def do_save():
    global save_timer
    with save_lock:
        if save_timer:
            save_timer.cancel()
            save_timer = None
        try:
            with open(_yaml_file_path, 'w') as config_file:
                _yaml.dump(_config, config_file, default_flow_style=None, width=150)
            if logger.isEnabledFor(logging.INFO):
                logger.info('saved %s to %s', _config, _yaml_file_path)
        except Exception as e:
            logger.error('failed to save to %s: %s', _yaml_file_path, e)


def _convert_json(json_dict):
    config = [json_dict.get(_KEY_VERSION)]
    for key, dev in json_dict.items():
        key = key.split(':')
        if len(key) == 2:
            dev[_KEY_WPID] = dev.get(_KEY_WPID) if dev.get(_KEY_WPID) else key[0]
            dev[_KEY_SERIAL] = dev.get(_KEY_SERIAL) if dev.get(_KEY_SERIAL) else key[1]
            for k, v in dev.items():
                if type(k) == str and not k.startswith('_') and type(v) == dict:  # convert string keys to ints
                    v = {int(dk) if type(dk) == str else dk: dv for dk, dv in v.items()}
                dev[k] = v
            for k in ['mouse-gestures', 'dpi-sliding']:
                v = dev.get(k, None)
                if v is True or v is False:
                    dev.pop(k)
            if '_name' in dev:
                dev[_KEY_NAME] = dev['_name']
                dev.pop('_name')
            config.append(dev)
    return config


class _DeviceEntry(dict):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        save(defer=True)

    def update(self, device, modelId):
        if device.name and device.name != self.get(_KEY_NAME):
            super().__setitem__(_KEY_NAME, device.name)
        if device.wpid and device.wpid != self.get(_KEY_WPID):
            super().__setitem__(_KEY_WPID, device.wpid)
        if device.serial and device.serial != '?' and device.serial != self.get(_KEY_SERIAL):
            super().__setitem__(_KEY_SERIAL, device.serial)
        if modelId and modelId != self.get(_KEY_MODEL_ID):
            super().__setitem__(_KEY_MODEL_ID, modelId)
        if device.unitId and device.unitId != self.get(_KEY_UNIT_ID):
            super().__setitem__(_KEY_UNIT_ID, device.unitId)

    def get_sensitivity(self, name):
        return self.get(_KEY_SENSITIVE, {}).get(name, False)

    def set_sensitivity(self, name, value):
        sensitives = self.get(_KEY_SENSITIVE, {})
        if sensitives.get(name) != value:
            sensitives[name] = value
            self.__setitem__(_KEY_SENSITIVE, sensitives)


def device_representer(dumper, data):
    return dumper.represent_mapping('tag:yaml.org,2002:map', data)


_yaml.add_representer(_DeviceEntry, device_representer)


def named_int_representer(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:int', str(int(data)))


_yaml.add_representer(_NamedInt, named_int_representer)


# A device can be identified by a combination of WPID and serial number (for receiver-connected devices)
# or a combination of modelId and unitId (for direct-connected devices).
# But some devices have empty (all zero) modelIds and unitIds.  Use the device name as a backup for the modelId.
# The worst situation is a receiver-connected device that Solaar has never seen on-line
# that is directly connected.  Here there is no way to realize that the two devices are the same.
# So new entries are not created for unseen off-line receiver-connected devices except for those with protocol 1.0
def persister(device):

    def match(wpid, serial, modelId, unitId, c):
        return ((wpid and wpid == c.get(_KEY_WPID) and serial and serial == c.get(_KEY_SERIAL)) or (
            modelId and modelId != '000000000000' and modelId == c.get(_KEY_MODEL_ID) and unitId
            and unitId == c.get(_KEY_UNIT_ID)
        ))

    if not _config:
        _load()
    entry = None
    modelId = device.modelId if device.modelId != '000000000000' else device.name if device.modelId else None
    for c in _config:
        if isinstance(c, _DeviceEntry) and match(device.wpid, device.serial, modelId, device.unitId, c):
            entry = c
            break
    if not entry:
        if not device.online and not device.serial:  # don't create entry for offline devices without serial number
            if logger.isEnabledFor(logging.INFO):
                logger.info('not setting up persister for offline device %s with missing serial number', device.name)
            return
        if logger.isEnabledFor(logging.INFO):
            logger.info('setting up persister for device %s', device.name)
        entry = _DeviceEntry()
        _config.append(entry)
    entry.update(device, modelId)
    return entry


def attach_to(device):
    pass



# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Translation support for the Logitech receivers library

import gettext as _gettext

_ = _gettext.gettext
ngettext = _gettext.ngettext

# A few common strings, not always accessible as such in the code.

_DUMMY = (
    # approximative battery levels
    _('empty'),
    _('critical'),
    _('low'),
    _('average'),
    _('good'),
    _('full'),

    # battery charging statuses
    _('discharging'),
    _('recharging'),
    _('charging'),
    _('not charging'),
    _('almost full'),
    _('charged'),
    _('slow recharge'),
    _('invalid battery'),
    _('thermal error'),
    _('error'),
    _('standard'),
    _('fast'),
    _('slow'),

    # pairing errors
    _('device timeout'),
    _('device not supported'),
    _('too many devices'),
    _('sequence timeout'),

    # firmware kinds
    _('Firmware'),
    _('Bootloader'),
    _('Hardware'),
    _('Other'),

    # common button and task names (from special_keys.py)
    _('Left Button'),
    _('Right Button'),
    _('Middle Button'),
    _('Back Button'),
    _('Forward Button'),
    _('Mouse Gesture Button'),
    _('Smart Shift'),
    _('DPI Switch'),
    _('Left Tilt'),
    _('Right Tilt'),
    _('Left Click'),
    _('Right Click'),
    _('Mouse Middle Button'),
    _('Mouse Back Button'),
    _('Mouse Forward Button'),
    _('Gesture Button Navigation'),
    _('Mouse Scroll Left Button'),
    _('Mouse Scroll Right Button'),

    # key/button statuses
    _('pressed'),
    _('released'),
)



# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Base low-level functions used by the API proper.
# Unlikely to be used directly unless you're expanding the API.

import logging
import threading as _threading

from collections import namedtuple
from contextlib import contextmanager
from random import getrandbits as _random_bits
from struct import pack as _pack
from time import time as _timestamp

import hidapi as _hid

from . import hidpp10 as _hidpp10
from . import hidpp20 as _hidpp20
from .base_usb import ALL as _RECEIVER_USB_IDS
from .base_usb import DEVICES as _DEVICE_IDS
from .base_usb import other_device_check as _other_device_check
from .common import KwException as _KwException
from .common import strhex as _strhex

logger = logging.getLogger(__name__)

#
#
#

_SHORT_MESSAGE_SIZE = 7
_LONG_MESSAGE_SIZE = 20
_MEDIUM_MESSAGE_SIZE = 15
_MAX_READ_SIZE = 32

HIDPP_SHORT_MESSAGE_ID = 0x10
HIDPP_LONG_MESSAGE_ID = 0x11
DJ_MESSAGE_ID = 0x20

# mapping from report_id to message length
report_lengths = {
    HIDPP_SHORT_MESSAGE_ID: _SHORT_MESSAGE_SIZE,
    HIDPP_LONG_MESSAGE_ID: _LONG_MESSAGE_SIZE,
    DJ_MESSAGE_ID: _MEDIUM_MESSAGE_SIZE,
    0x21: _MAX_READ_SIZE
}
"""Default timeout on read (in seconds)."""
DEFAULT_TIMEOUT = 4
# the receiver itself should reply very fast, within 500ms
_RECEIVER_REQUEST_TIMEOUT = 0.9
# devices may reply a lot slower, as the call has to go wireless to them and come back
_DEVICE_REQUEST_TIMEOUT = DEFAULT_TIMEOUT
# when pinging, be extra patient (no longer)
_PING_TIMEOUT = DEFAULT_TIMEOUT

#
# Exceptions that may be raised by this API.
#


class NoReceiver(_KwException):
    """Raised when trying to talk through a previously open handle, when the
    receiver is no longer available. Should only happen if the receiver is
    physically disconnected from the machine, or its kernel driver module is
    unloaded."""
    pass


class NoSuchDevice(_KwException):
    """Raised when trying to reach a device number not paired to the receiver."""
    pass


class DeviceUnreachable(_KwException):
    """Raised when a request is made to an unreachable (turned off) device."""
    pass


#
#
#


def match(record, bus_id, vendor_id, product_id):
    return ((record.get('bus_id') is None or record.get('bus_id') == bus_id)
            and (record.get('vendor_id') is None or record.get('vendor_id') == vendor_id)
            and (record.get('product_id') is None or record.get('product_id') == product_id))


def filter_receivers(bus_id, vendor_id, product_id, hidpp_short=False, hidpp_long=False):
    """Check that this product is a Logitech receiver and if so return the receiver record for further checking"""
    for record in _RECEIVER_USB_IDS:  # known receivers
        if match(record, bus_id, vendor_id, product_id):
            return record
    if vendor_id == 0x046D and 0xC500 <= product_id <= 0xC5FF:  # unknown receiver
        return {'vendor_id': vendor_id, 'product_id': product_id, 'bus_id': bus_id, 'isDevice': False}


def receivers():
    """Enumerate all the receivers attached to the machine."""
    yield from _hid.enumerate(filter_receivers)


def filter(bus_id, vendor_id, product_id, hidpp_short=False, hidpp_long=False):
    """Check that this product is of interest and if so return the device record for further checking"""
    record = filter_receivers(bus_id, vendor_id, product_id, hidpp_short, hidpp_long)
    if record:  # known or unknown receiver
        return record
    for record in _DEVICE_IDS:  # known devices
        if match(record, bus_id, vendor_id, product_id):
            return record
    if hidpp_short or hidpp_long:  # unknown devices that use HID++
        return {'vendor_id': vendor_id, 'product_id': product_id, 'bus_id': bus_id, 'isDevice': True}
    elif hidpp_short is None and hidpp_long is None:  # unknown devices in correct range of IDs
        return _other_device_check(bus_id, vendor_id, product_id)


def receivers_and_devices():
    """Enumerate all the receivers and devices directly attached to the machine."""
    yield from _hid.enumerate(filter)


def notify_on_receivers_glib(callback):
    """Watch for matching devices and notifies the callback on the GLib thread."""
    return _hid.monitor_glib(callback, filter)


#
#
#


def open_path(path):
    """Checks if the given Linux device path points to the right UR device.

    :param path: the Linux device path.

    The UR physical device may expose multiple linux devices with the same
    interface, so we have to check for the right one. At this moment the only
    way to distinguish betheen them is to do a test ping on an invalid
    (attached) device number (i.e., 0), expecting a 'ping failed' reply.

    :returns: an open receiver handle if this is the right Linux device, or
    ``None``.
    """
    return _hid.open_path(path)


def open():
    """Opens the first Logitech Unifying Receiver found attached to the machine.

    :returns: An open file handle for the found receiver, or ``None``.
    """
    for rawdevice in receivers():
        handle = open_path(rawdevice.path)
        if handle:
            return handle


def close(handle):
    """Closes a HID device handle."""
    if handle:
        try:
            if isinstance(handle, int):
                _hid.close(handle)
            else:
                handle.close()
            # logger.info("closed receiver handle %r", handle)
            return True
        except Exception:
            # logger.exception("closing receiver handle %r", handle)
            pass

    return False


def write(handle, devnumber, data, long_message=False):
    """Writes some data to the receiver, addressed to a certain device.

    :param handle: an open UR handle.
    :param devnumber: attached device number.
    :param data: data to send, up to 5 bytes.

    The first two (required) bytes of data must be the SubId and address.

    :raises NoReceiver: if the receiver is no longer available, i.e. has
    been physically removed from the machine, or the kernel driver has been
    unloaded. The handle will be closed automatically.
    """
    # the data is padded to either 5 or 18 bytes
    assert data is not None
    assert isinstance(data, bytes), (repr(data), type(data))

    if long_message or len(data) > _SHORT_MESSAGE_SIZE - 2 or data[:1] == b'\x82':
        wdata = _pack('!BB18s', HIDPP_LONG_MESSAGE_ID, devnumber, data)
    else:
        wdata = _pack('!BB5s', HIDPP_SHORT_MESSAGE_ID, devnumber, data)
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('(%s) <= w[%02X %02X %s %s]', handle, ord(wdata[:1]), devnumber, _strhex(wdata[2:4]), _strhex(wdata[4:]))

    try:
        _hid.write(int(handle), wdata)
    except Exception as reason:
        logger.error('write failed, assuming handle %r no longer available', handle)
        close(handle)
        raise NoReceiver(reason=reason)


def read(handle, timeout=DEFAULT_TIMEOUT):
    """Read some data from the receiver. Usually called after a write (feature
    call), to get the reply.

    :param: handle open handle to the receiver
    :param: timeout how long to wait for a reply, in seconds

    :returns: a tuple of (devnumber, message data), or `None`

    :raises NoReceiver: if the receiver is no longer available, i.e. has
    been physically removed from the machine, or the kernel driver has been
    unloaded. The handle will be closed automatically.
    """
    reply = _read(handle, timeout)
    if reply:
        return reply


# sanity checks on  message report id and size
def check_message(data):
    assert isinstance(data, bytes), (repr(data), type(data))
    report_id = ord(data[:1])
    if report_id in report_lengths:  # is this an HID++ or DJ message?
        if report_lengths.get(report_id) == len(data):
            return True
        else:
            logger.warning('unexpected message size: report_id %02X message %s' % (report_id, _strhex(data)))
    return False


def _read(handle, timeout):
    """Read an incoming packet from the receiver.

    :returns: a tuple of (report_id, devnumber, data), or `None`.

    :raises NoReceiver: if the receiver is no longer available, i.e. has
    been physically removed from the machine, or the kernel driver has been
    unloaded. The handle will be closed automatically.
    """
    try:
        # convert timeout to milliseconds, the hidapi expects it
        timeout = int(timeout * 1000)
        data = _hid.read(int(handle), _MAX_READ_SIZE, timeout)
    except Exception as reason:
        logger.warning('read failed, assuming handle %r no longer available', handle)
        close(handle)
        raise NoReceiver(reason=reason)

    if data and check_message(data):  # ignore messages that fail check
        report_id = ord(data[:1])
        devnumber = ord(data[1:2])

        if logger.isEnabledFor(logging.DEBUG
                               ) and (report_id != DJ_MESSAGE_ID or ord(data[2:3]) > 0x10):  # ignore DJ input messages
            logger.debug('(%s) => r[%02X %02X %s %s]', handle, report_id, devnumber, _strhex(data[2:4]), _strhex(data[4:]))

        return report_id, devnumber, data[2:]


#
#
#


def _skip_incoming(handle, ihandle, notifications_hook):
    """Read anything already in the input buffer.

    Used by request() and ping() before their write.
    """

    while True:
        try:
            # read whatever is already in the buffer, if any
            data = _hid.read(ihandle, _MAX_READ_SIZE, 0)
        except Exception as reason:
            logger.error('read failed, assuming receiver %s no longer available', handle)
            close(handle)
            raise NoReceiver(reason=reason)

        if data:
            if check_message(data):  # only process messages that pass check
                # report_id = ord(data[:1])
                if notifications_hook:
                    n = make_notification(ord(data[:1]), ord(data[1:2]), data[2:])
                    if n:
                        notifications_hook(n)
        else:
            # nothing in the input buffer, we're done
            return


def make_notification(report_id, devnumber, data):
    """Guess if this is a notification (and not just a request reply), and
    return a Notification tuple if it is."""

    sub_id = ord(data[:1])
    if sub_id & 0x80 == 0x80:
        # this is either a HID++1.0 register r/w, or an error reply
        return

    # DJ input records are not notifications
    if report_id == DJ_MESSAGE_ID and (sub_id < 0x10):
        return

    address = ord(data[1:2])
    if sub_id == 0x00 and (address & 0x0F == 0x00):
        # this is a no-op notification - don't do anything with it
        return

    if (
        # standard HID++ 1.0 notification, SubId may be 0x40 - 0x7F
        (sub_id >= 0x40) or  # noqa: E131
        # custom HID++1.0 battery events, where SubId is 0x07/0x0D
        (sub_id in (0x07, 0x0D) and len(data) == 5 and data[4:5] == b'\x00') or
        # custom HID++1.0 illumination event, where SubId is 0x17
        (sub_id == 0x17 and len(data) == 5) or
        # HID++ 2.0 feature notifications have the SoftwareID 0
        (address & 0x0F == 0x00)
    ):  # noqa: E129
        return _HIDPP_Notification(report_id, devnumber, sub_id, address, data[2:])


_HIDPP_Notification = namedtuple('_HIDPP_Notification', ('report_id', 'devnumber', 'sub_id', 'address', 'data'))
_HIDPP_Notification.__str__ = lambda self: 'Notification(%02x,%d,%02X,%02X,%s)' % (
    self.report_id, self.devnumber, self.sub_id, self.address, _strhex(self.data)
)
del namedtuple

#
#
#

request_lock = _threading.Lock()  # serialize all requests
handles_lock = {}


def handle_lock(handle):
    with request_lock:
        if handles_lock.get(handle) is None:
            if logger.isEnabledFor(logging.INFO):
                logger.info('New lock %s', repr(handle))
            handles_lock[handle] = _threading.Lock()  # Serialize requests on the handle
    return handles_lock[handle]


# context manager for locks with a timeout
@contextmanager
def acquire_timeout(lock, handle, timeout):
    result = lock.acquire(timeout=timeout)
    try:
        if not result:
            logger.error('lock on handle %d not acquired, probably due to timeout', int(handle))
        yield result
    finally:
        if result:
            lock.release()


# a very few requests (e.g., host switching) do not expect a reply, but use no_reply=True with extreme caution
def request(handle, devnumber, request_id, *params, no_reply=False, return_error=False, long_message=False, protocol=1.0):
    """Makes a feature call to a device and waits for a matching reply.
    :param handle: an open UR handle.
    :param devnumber: attached device number.
    :param request_id: a 16-bit integer.
    :param params: parameters for the feature call, 3 to 16 bytes.
    :returns: the reply data, or ``None`` if some error occurred. or no reply expected
    """

    # import inspect as _inspect
    # print ('\n  '.join(str(s) for s in _inspect.stack()))

    with acquire_timeout(handle_lock(handle), handle, 10.):
        assert isinstance(request_id, int)
        if (devnumber != 0xFF or protocol >= 2.0) and request_id < 0x8000:
            # For HID++ 2.0 feature requests, randomize the SoftwareId to make it
            # easier to recognize the reply for this request. also, always set the
            # most significant bit (8) in SoftwareId, to make notifications easier
            # to distinguish from request replies.
            # This only applies to peripheral requests, ofc.
            request_id = (request_id & 0xFFF0) | 0x08 | _random_bits(3)

        timeout = _RECEIVER_REQUEST_TIMEOUT if devnumber == 0xFF else _DEVICE_REQUEST_TIMEOUT
        # be extra patient on long register read
        if request_id & 0xFF00 == 0x8300:
            timeout *= 2

        if params:
            params = b''.join(_pack('B', p) if isinstance(p, int) else p for p in params)
        else:
            params = b''
        # if logger.isEnabledFor(logging.DEBUG):
        #     logger.debug("(%s) device %d request_id {%04X} params [%s]", handle, devnumber, request_id, _strhex(params))
        request_data = _pack('!H', request_id) + params

        ihandle = int(handle)
        notifications_hook = getattr(handle, 'notifications_hook', None)
        try:
            _skip_incoming(handle, ihandle, notifications_hook)
        except NoReceiver:
            logger.warning('device or receiver disconnected')
            return None
        write(ihandle, devnumber, request_data, long_message)

        if no_reply:
            return None

        # we consider timeout from this point
        request_started = _timestamp()
        delta = 0

        while delta < timeout:
            reply = _read(handle, timeout)

            if reply:
                report_id, reply_devnumber, reply_data = reply
                if reply_devnumber == devnumber or reply_devnumber == devnumber ^ 0xff:  # BT device returning 0x00
                    if report_id == HIDPP_SHORT_MESSAGE_ID and reply_data[:1] == b'\x8F' and reply_data[1:3] == request_data[:2
                                                                                                                             ]:
                        error = ord(reply_data[3:4])

                        if logger.isEnabledFor(logging.DEBUG):
                            logger.debug(
                                '(%s) device 0x%02X error on request {%04X}: %d = %s', handle, devnumber, request_id, error,
                                _hidpp10.ERROR[error]
                            )
                        return _hidpp10.ERROR[error] if return_error else None
                    if reply_data[:1] == b'\xFF' and reply_data[1:3] == request_data[:2]:
                        # a HID++ 2.0 feature call returned with an error
                        error = ord(reply_data[3:4])
                        logger.error(
                            '(%s) device %d error on feature request {%04X}: %d = %s', handle, devnumber, request_id, error,
                            _hidpp20.ERROR[error]
                        )
                        raise _hidpp20.FeatureCallError(number=devnumber, request=request_id, error=error, params=params)

                    if reply_data[:2] == request_data[:2]:
                        if devnumber == 0xFF:
                            if request_id == 0x83B5 or request_id == 0x81F1:
                                # these replies have to match the first parameter as well
                                if reply_data[2:3] == params[:1]:
                                    return reply_data[2:]
                                else:
                                    # hm, not matching my request, and certainly not a notification
                                    continue
                            else:
                                return reply_data[2:]
                        else:
                            return reply_data[2:]
                else:
                    # a reply was received, but did not match our request in any way
                    # reset the timeout starting point
                    request_started = _timestamp()

                if notifications_hook:
                    n = make_notification(report_id, reply_devnumber, reply_data)
                    if n:
                        notifications_hook(n)
                    # elif logger.isEnabledFor(logging.DEBUG):
                    #     logger.debug("(%s) ignoring reply %02X [%s]", handle, reply_devnumber, _strhex(reply_data))
                # elif logger.isEnabledFor(logging.DEBUG):
                #     logger.debug("(%s) ignoring reply %02X [%s]", handle, reply_devnumber, _strhex(reply_data))

            delta = _timestamp() - request_started
            # if logger.isEnabledFor(logging.DEBUG):
            #     logger.debug("(%s) still waiting for reply, delta %f", handle, delta)

        logger.warning(
            'timeout (%0.2f/%0.2f) on device %d request {%04X} params [%s]', delta, timeout, devnumber, request_id,
            _strhex(params)
        )
        # raise DeviceUnreachable(number=devnumber, request=request_id)


def ping(handle, devnumber, long_message=False):
    """Check if a device is connected to the receiver.
    :returns: The HID protocol supported by the device, as a floating point number, if the device is active.
    """
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('(%s) pinging device %d', handle, devnumber)
    with acquire_timeout(handle_lock(handle), handle, 10.):
        notifications_hook = getattr(handle, 'notifications_hook', None)
        try:
            _skip_incoming(handle, int(handle), notifications_hook)
        except NoReceiver:
            logger.warning('device or receiver disconnected')
            return

        # randomize the SoftwareId and mark byte to be able to identify the ping
        # reply, and set most significant (0x8) bit in SoftwareId so that the reply
        # is always distinguishable from notifications
        request_id = 0x0018 | _random_bits(3)
        request_data = _pack('!HBBB', request_id, 0, 0, _random_bits(8))
        write(int(handle), devnumber, request_data, long_message)

        request_started = _timestamp()  # we consider timeout from this point
        delta = 0
        while delta < _PING_TIMEOUT:
            reply = _read(handle, _PING_TIMEOUT)
            if reply:
                report_id, reply_devnumber, reply_data = reply
                if reply_devnumber == devnumber or reply_devnumber == devnumber ^ 0xff:  # BT device returning 0x00
                    if reply_data[:2] == request_data[:2] and reply_data[4:5] == request_data[-1:]:
                        # HID++ 2.0+ device, currently connected
                        return ord(reply_data[2:3]) + ord(reply_data[3:4]) / 10.0

                    if report_id == HIDPP_SHORT_MESSAGE_ID and reply_data[:1] == b'\x8F' and \
                       reply_data[1:3] == request_data[:2]:  # error response
                        error = ord(reply_data[3:4])
                        if error == _hidpp10.ERROR.invalid_SubID__command:  # a valid reply from a HID++ 1.0 device
                            return 1.0
                        if error == _hidpp10.ERROR.resource_error or error == _hidpp10.ERROR.connection_request_failed:
                            return  # device unreachable
                        if error == _hidpp10.ERROR.unknown_device:  # no paired device with that number
                            logger.error('(%s) device %d error on ping request: unknown device', handle, devnumber)
                            raise NoSuchDevice(number=devnumber, request=request_id)

                if notifications_hook:
                    n = make_notification(report_id, reply_devnumber, reply_data)
                    if n:
                        notifications_hook(n)
                    # elif logger.isEnabledFor(logging.DEBUG):
                    #     logger.debug("(%s) ignoring reply %02X [%s]", handle, reply_devnumber, _strhex(reply_data))

            delta = _timestamp() - request_started

        logger.warning('(%s) timeout (%0.2f/%0.2f) on device %d ping', handle, delta, _PING_TIMEOUT, devnumber)



# -*- python-mode -*-

## Copyright (C) 2020
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from logitech_receiver import base as _base
from logitech_receiver import hidpp10 as _hidpp10
from logitech_receiver.common import strhex as _strhex
from solaar.cli.show import _print_device, _print_receiver

_R = _hidpp10.REGISTERS


def run(receivers, args, find_receiver, _ignore):
    assert receivers

    if args.receiver:
        receiver_name = args.receiver.lower()
        receiver = find_receiver(receivers, receiver_name)
        if not receiver:
            raise Exception("no receiver found matching '%s'" % receiver_name)
    else:
        receiver = receivers[0]

    assert receiver is not None

    if receiver.isDevice:
        _print_device(receiver, 1)
        return

    _print_receiver(receiver)

    print('')
    print('  Register Dump')
    rgst = receiver.read_register(_R.notifications)
    print('    Notifications         %#04x: %s' % (_R.notifications % 0x100, '0x' + _strhex(rgst) if rgst else 'None'))
    rgst = receiver.read_register(_R.receiver_connection)
    print('    Connection State      %#04x: %s' % (_R.receiver_connection % 0x100, '0x' + _strhex(rgst) if rgst else 'None'))
    rgst = receiver.read_register(_R.devices_activity)
    print('    Device Activity       %#04x: %s' % (_R.devices_activity % 0x100, '0x' + _strhex(rgst) if rgst else 'None'))

    for sub_reg in range(0, 16):
        rgst = receiver.read_register(_R.receiver_info, sub_reg)
        print(
            '    Pairing Register %#04x %#04x: %s' %
            (_R.receiver_info % 0x100, sub_reg, '0x' + _strhex(rgst) if rgst else 'None')
        )
    for device in range(0, 7):
        for sub_reg in [0x10, 0x20, 0x30, 0x50]:
            rgst = receiver.read_register(_R.receiver_info, sub_reg + device)
            print(
                '    Pairing Register %#04x %#04x: %s' %
                (_R.receiver_info % 0x100, sub_reg + device, '0x' + _strhex(rgst) if rgst else 'None')
            )
        rgst = receiver.read_register(_R.receiver_info, 0x40 + device)
        print(
            '    Pairing Name     %#04x %#02x: %s' %
            (_R.receiver_info % 0x100, 0x40 + device, rgst[2:2 + ord(rgst[1:2])] if rgst else 'None')
        )
        for part in range(1, 4):
            rgst = receiver.read_register(_R.receiver_info, 0x60 + device, part)
            print(
                '    Pairing Name     %#04x %#02x %#02x: %2d %s' % (
                    _R.receiver_info % 0x100, 0x60 + device, part, ord(rgst[2:3]) if rgst else 0,
                    rgst[3:3 + ord(rgst[2:3])] if rgst else 'None'
                )
            )
    for sub_reg in range(0, 5):
        rgst = receiver.read_register(_R.firmware, sub_reg)
        print(
            '    Firmware         %#04x %#04x: %s' %
            (_R.firmware % 0x100, sub_reg, '0x' + _strhex(rgst) if rgst is not None else 'None')
        )

    print('')
    for reg in range(0, 0xFF):
        last = None
        for sub in range(0, 0xFF):
            rgst = _base.request(receiver.handle, 0xFF, 0x8100 | reg, sub, return_error=True)
            if isinstance(rgst, int) and rgst == _hidpp10.ERROR.invalid_address:
                break
            elif isinstance(rgst, int) and rgst == _hidpp10.ERROR.invalid_value:
                continue
            else:
                if not isinstance(last, bytes) or not isinstance(rgst, bytes) or last != rgst:
                    print(
                        '    Register Short   %#04x %#04x: %s' %
                        (reg, sub, '0x' + _strhex(rgst) if isinstance(rgst, bytes) else str(rgst))
                    )
            last = rgst
        last = None
        for sub in range(0, 0xFF):
            rgst = _base.request(receiver.handle, 0xFF, 0x8100 | (0x200 + reg), sub, return_error=True)
            if isinstance(rgst, int) and rgst == _hidpp10.ERROR.invalid_address:
                break
            elif isinstance(rgst, int) and rgst == _hidpp10.ERROR.invalid_value:
                continue
            else:
                if not isinstance(last, bytes) or not isinstance(rgst, bytes) or last != rgst:
                    print(
                        '    Register Long    %#04x %#04x: %s' %
                        (reg, sub, '0x' + _strhex(rgst) if isinstance(rgst, bytes) else str(rgst))
                    )
            last = rgst



# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import pkgutil as _pkgutil
import subprocess as _subprocess
import sys as _sys

NAME = 'solaar'

try:
    __version__ = _subprocess.check_output(['git', 'describe', '--always'], cwd=_sys.path[0],
                                           stderr=_subprocess.DEVNULL).strip().decode()
except Exception:
    try:
        __version__ = _pkgutil.get_data('solaar', 'commit').strip().decode()
    except Exception:
        __version__ = _pkgutil.get_data('solaar', 'version').strip().decode()



#!/usr/bin/env python3
# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import importlib
import logging
import os.path
import platform
import signal
import sys
import tempfile

import solaar.cli as _cli
import solaar.i18n as _i18n

from solaar import NAME, __version__

logger = logging.getLogger(__name__)

#
#
#


def _require(module, os_package, gi=None, gi_package=None, gi_version=None):
    try:
        if gi is not None:
            gi.require_version(gi_package, gi_version)
        return importlib.import_module(module)
    except (ImportError, ValueError):
        sys.exit('%s: missing required system package %s' % (NAME, os_package))


battery_icons_style = 'regular'
temp = tempfile.NamedTemporaryFile(prefix='Solaar_', mode='w', delete=True)


def _parse_arguments():
    import argparse
    arg_parser = argparse.ArgumentParser(
        prog=NAME.lower(), epilog='For more information see https://pwr-solaar.github.io/Solaar'
    )
    arg_parser.add_argument(
        '-d',
        '--debug',
        action='count',
        default=0,
        help='print logging messages, for debugging purposes (may be repeated for extra verbosity)'
    )
    arg_parser.add_argument(
        '-D',
        '--hidraw',
        action='store',
        dest='hidraw_path',
        metavar='PATH',
        help='unifying receiver to use; the first detected receiver if unspecified. Example: /dev/hidraw2'
    )
    arg_parser.add_argument('--restart-on-wake-up', action='store_true', help='restart Solaar on sleep wake-up (experimental)')
    arg_parser.add_argument(
        '-w', '--window', choices=('show', 'hide', 'only'), help='start with window showing / hidden / only (no tray icon)'
    )
    arg_parser.add_argument(
        '-b',
        '--battery-icons',
        choices=('regular', 'symbolic', 'solaar'),
        help='prefer regular battery / symbolic battery / solaar icons'
    )
    arg_parser.add_argument('--tray-icon-size', type=int, help='explicit size for tray icons')
    arg_parser.add_argument('-V', '--version', action='version', version='%(prog)s ' + __version__)
    arg_parser.add_argument('--help-actions', action='store_true', help='print help for the optional actions')
    arg_parser.add_argument('action', nargs=argparse.REMAINDER, choices=_cli.actions, help='optional actions to perform')

    args = arg_parser.parse_args()

    if args.help_actions:
        _cli.print_help()
        return

    if args.window is None:
        args.window = 'show'  # default behaviour is to show main window

    global battery_icons_style
    battery_icons_style = args.battery_icons if args.battery_icons is not None else 'regular'
    global tray_icon_size
    tray_icon_size = args.tray_icon_size

    log_format = '%(asctime)s,%(msecs)03d %(levelname)8s [%(threadName)s] %(name)s: %(message)s'
    log_level = logging.ERROR - 10 * args.debug
    logging.getLogger('').setLevel(min(log_level, logging.WARNING))
    file_handler = logging.StreamHandler(temp)
    file_handler.setLevel(max(min(log_level, logging.WARNING), logging.INFO))
    file_handler.setFormatter(logging.Formatter(log_format))
    logging.getLogger('').addHandler(file_handler)
    if args.debug > 0:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter(log_format))
        stream_handler.setLevel(log_level)
        logging.getLogger('').addHandler(stream_handler)

    if not args.action:
        if logger.isEnabledFor(logging.INFO):
            logger.info('version %s, language %s (%s)', __version__, _i18n.language, _i18n.encoding)

    return args


# On first SIGINT, dump threads to stderr; on second, exit
def _handlesig(signl, stack):
    import faulthandler
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGTERM, signal.SIG_DFL)

    if signl == int(signal.SIGINT):
        if logger.isEnabledFor(logging.INFO):
            faulthandler.dump_traceback()
        sys.exit('%s: exit due to keyboard interrupt' % (NAME.lower()))
    else:
        sys.exit(0)


def main():
    if platform.system() not in ('Darwin', 'Windows'):
        _require('pyudev', 'python3-pyudev')

    args = _parse_arguments()
    if not args:
        return
    if args.action:
        # if any argument, run comandline and exit
        return _cli.run(args.action, args.hidraw_path)

    gi = _require('gi', 'python3-gi (in Ubuntu) or python3-gobject (in Fedora)')
    _require('gi.repository.Gtk', 'gir1.2-gtk-3.0', gi, 'Gtk', '3.0')

    # handle ^C in console
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGINT, _handlesig)
    signal.signal(signal.SIGTERM, _handlesig)

    udev_file = '42-logitech-unify-permissions.rules'
    if logger.isEnabledFor(logging.WARNING) \
       and not os.path.isfile('/etc/udev/rules.d/' + udev_file) \
       and not os.path.isfile('/usr/lib/udev/rules.d/' + udev_file) \
       and not os.path.isfile('/usr/local/lib/udev/rules.d/' + udev_file):
        logger.warning('Solaar udev file not found in expected location')
        logger.warning('See https://pwr-solaar.github.io/Solaar/installation for more information')
    try:
        import solaar.listener as listener
        import solaar.ui as ui

        listener.setup_scanner(ui.status_changed, ui.error_dialog)

        import solaar.upower as _upower
        if args.restart_on_wake_up:
            _upower.watch(listener.start_all, listener.stop_all)
        else:
            _upower.watch(lambda: listener.ping_all(True))

        import solaar.configuration as _configuration
        _configuration.defer_saves = True  # allow configuration saves to be deferred

        # main UI event loop
        ui.run_loop(listener.start_all, listener.stop_all, args.window != 'only', args.window != 'hide')
    except Exception:
        from traceback import format_exc
        sys.exit('%s: error: %s' % (NAME.lower(), format_exc()))

    temp.close()


if __name__ == '__main__':
    main()



# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging

import solaar.gtk as gtk

from gi.repository import Gtk

logger = logging.getLogger(__name__)

#
#
#

_LARGE_SIZE = 64
Gtk.IconSize.LARGE = Gtk.icon_size_register('large', _LARGE_SIZE, _LARGE_SIZE)
# Gtk.IconSize.XLARGE = Gtk.icon_size_register('x-large', _LARGE_SIZE * 2, _LARGE_SIZE * 2)

TRAY_INIT = 'solaar-init'
TRAY_OKAY = 'solaar'
TRAY_ATTENTION = 'solaar-attention'

_default_theme = None


def _init_icon_paths():
    global _default_theme
    if _default_theme:
        return

    _default_theme = Gtk.IconTheme.get_default()
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('icon theme paths: %s', _default_theme.get_search_path())

    if gtk.battery_icons_style == 'symbolic':
        global TRAY_OKAY
        TRAY_OKAY = TRAY_INIT  # use monochrome tray icon
        if not _default_theme.has_icon('battery-good-symbolic'):
            logger.warning('failed to detect symbolic icons')
            gtk.battery_icons_style = 'regular'
    if gtk.battery_icons_style == 'regular':
        if not _default_theme.has_icon('battery-good'):
            logger.warning('failed to detect icons')
            gtk.battery_icons_style = 'solaar'


#
#
#


def battery(level=None, charging=False):
    icon_name = _battery_icon_name(level, charging)
    if not _default_theme.has_icon(icon_name):
        logger.warning('icon %s not found in current theme', icon_name)
        return TRAY_OKAY  # use Solaar icon if battery icon not available
    elif logger.isEnabledFor(logging.DEBUG):
        logger.debug('battery icon for %s:%s = %s', level, charging, icon_name)
    return icon_name


# return first res where val >= guard
# _first_res(val,((guard,res),...))
def _first_res(val, pairs):
    return next((res for guard, res in pairs if val >= guard), None)


def _battery_icon_name(level, charging):
    _init_icon_paths()

    if level is None or level < 0:
        return 'battery-missing' + ('-symbolic' if gtk.battery_icons_style == 'symbolic' else '')

    level_name = _first_res(level, ((90, 'full'), (30, 'good'), (20, 'low'), (5, 'caution'), (0, 'empty')))
    return 'battery-%s%s%s' % (
        level_name, '-charging' if charging else '', '-symbolic' if gtk.battery_icons_style == 'symbolic' else ''
    )


#
#
#


def lux(level=None):
    if level is None or level < 0:
        return 'light_unknown'
    return 'solaar-light_%03d' % (20 * ((level + 50) // 100))


#
#
#

_ICON_SETS = {}


def device_icon_set(name='_', kind=None):
    icon_set = _ICON_SETS.get(name)
    if icon_set is None:
        icon_set = Gtk.IconSet.new()
        _ICON_SETS[name] = icon_set

        # names of possible icons, in reverse order of likelihood
        # the theme will hopefully pick up the most appropriate
        names = ['preferences-desktop-peripherals']
        if kind:
            if str(kind) == 'numpad':
                names += ('input-keyboard', 'input-dialpad')
            elif str(kind) == 'touchpad':
                names += ('input-mouse', 'input-tablet')
            elif str(kind) == 'trackball':
                names += ('input-mouse', )
            elif str(kind) == 'headset':
                names += ('audio-headphones', 'audio-headset')
            names += ('input-' + str(kind), )
        # names += (name.replace(' ', '-'),)

        source = Gtk.IconSource.new()
        for n in names:
            source.set_icon_name(n)
            icon_set.add_source(source)
        icon_set.names = names

    return icon_set


def device_icon_file(name, kind=None, size=_LARGE_SIZE):
    _init_icon_paths()

    icon_set = device_icon_set(name, kind)
    assert icon_set
    for n in reversed(icon_set.names):
        if _default_theme.has_icon(n):
            return _default_theme.lookup_icon(n, size, 0).get_filename()


def device_icon_name(name, kind=None):
    _init_icon_paths()

    icon_set = device_icon_set(name, kind)
    assert icon_set
    for n in reversed(icon_set.names):
        if _default_theme.has_icon(n):
            return n


def icon_file(name, size=_LARGE_SIZE):
    _init_icon_paths()

    # has_icon() somehow returned False while lookup_icon returns non-None.
    # I guess it happens because share/solaar/icons/ has no hicolor and
    # resolution subdirs
    theme_icon = _default_theme.lookup_icon(name, size, 0)
    if theme_icon:
        file_name = theme_icon.get_filename()
        # if logger.isEnabledFor(logging.DEBUG):
        #     logger.debug("icon %s(%d) => %s", name, size, file_name)
        return file_name

    logger.warning('icon %s(%d) not found in current theme', name, size)



# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging

logger = logging.getLogger(__name__)

#
# As suggested here: http://stackoverflow.com/a/13548984
#

_suspend_callback = None


def _suspend():
    if _suspend_callback:
        if logger.isEnabledFor(logging.INFO):
            logger.info('received suspend event')
        _suspend_callback()


_resume_callback = None


def _resume():
    if _resume_callback:
        if logger.isEnabledFor(logging.INFO):
            logger.info('received resume event')
        _resume_callback()


def _suspend_or_resume(suspend):
    _suspend() if suspend else _resume()


def watch(on_resume_callback=None, on_suspend_callback=None):
    """Register callback for suspend/resume events.
    They are called only if the system DBus is running, and the Login daemon is available."""
    global _resume_callback, _suspend_callback
    _suspend_callback = on_suspend_callback
    _resume_callback = on_resume_callback


try:
    import dbus

    _LOGIND_BUS = 'org.freedesktop.login1'
    _LOGIND_INTERFACE = 'org.freedesktop.login1.Manager'

    # integration into the main GLib loop
    from dbus.mainloop.glib import DBusGMainLoop
    DBusGMainLoop(set_as_default=True)

    bus = dbus.SystemBus()
    assert bus

    bus.add_signal_receiver(_suspend_or_resume, 'PrepareForSleep', dbus_interface=_LOGIND_INTERFACE, bus_name=_LOGIND_BUS)

    if logger.isEnabledFor(logging.INFO):
        logger.info('connected to system dbus, watching for suspend/resume events')

except Exception:
    # Either:
    # - the dbus library is not available
    # - the system dbus is not running
    logger.warning('failed to register suspend/resume callbacks')
    pass



# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Reprogrammable keys information
# Mostly from Logitech documentation, but with some edits for better Lunix compatibility

from .common import NamedInts as _NamedInts
from .common import UnsortedNamedInts as _UnsortedNamedInts

# <controls.xml awk -F\" '/<Control /{sub(/^LD_FINFO_(CTRLID_)?/, "", $2);printf("\t%s=0x%04X,\n", $2, $4)}' | sort -t= -k2
CONTROL = _NamedInts({
    'Volume_Up': 0x0001,
    'Volume_Down': 0x0002,
    'Mute': 0x0003,
    'Play__Pause': 0x0004,
    'Next': 0x0005,
    'Previous': 0x0006,
    'Stop': 0x0007,
    'Application_Switcher': 0x0008,
    'Burn': 0x0009,
    'Calculator': 0x000A,  # Craft Keyboard top 4th from right
    'Calendar': 0x000B,
    'Close': 0x000C,
    'Eject': 0x000D,
    'Mail': 0x000E,
    'Help_As_HID': 0x000F,
    'Help_As_F1': 0x0010,
    'Launch_Word_Proc': 0x0011,
    'Launch_Spreadsheet': 0x0012,
    'Launch_Presentation': 0x0013,
    'Undo_As_Ctrl_Z': 0x0014,
    'Undo_As_HID': 0x0015,
    'Redo_As_Ctrl_Y': 0x0016,
    'Redo_As_HID': 0x0017,
    'Print_As_Ctrl_P': 0x0018,
    'Print_As_HID': 0x0019,
    'Save_As_Ctrl_S': 0x001A,
    'Save_As_HID': 0x001B,
    'Preset_A': 0x001C,
    'Preset_B': 0x001D,
    'Preset_C': 0x001E,
    'Preset_D': 0x001F,
    'Favorites': 0x0020,
    'Gadgets': 0x0021,
    'My_Home': 0x0022,
    'Gadgets_As_Win_G': 0x0023,
    'Maximize_As_HID': 0x0024,
    'Maximize_As_Win_Shift_M': 0x0025,
    'Minimize_As_HID': 0x0026,
    'Minimize_As_Win_M': 0x0027,
    'Media_Player': 0x0028,
    'Media_Center_Logi': 0x0029,
    'Media_Center_Msft': 0x002A,  # Should not be used as it is not reprogrammable under Windows
    'Custom_Menu': 0x002B,
    'Messenger': 0x002C,
    'My_Documents': 0x002D,
    'My_Music': 0x002E,
    'Webcam': 0x002F,
    'My_Pictures': 0x0030,
    'My_Videos': 0x0031,
    'My_Computer_As_HID': 0x0032,
    'My_Computer_As_Win_E': 0x0033,
    'FN_Key': 0x0034,
    'Launch_Picture_Viewer': 0x0035,
    'One_Touch_Search': 0x0036,
    'Preset_1': 0x0037,
    'Preset_2': 0x0038,
    'Preset_3': 0x0039,
    'Preset_4': 0x003A,
    'Record': 0x003B,
    'Internet_Refresh': 0x003C,
    'Search': 0x003E,  # SEARCH
    'Shuffle': 0x003F,
    'Sleep': 0x0040,
    'Internet_Stop': 0x0041,
    'Synchronize': 0x0042,
    'Zoom': 0x0043,
    'Zoom_In_As_HID': 0x0044,
    'Zoom_In_As_Ctrl_Wheel': 0x0045,
    'Zoom_In_As_Cltr_Plus': 0x0046,
    'Zoom_Out_As_HID': 0x0047,
    'Zoom_Out_As_Ctrl_Wheel': 0x0048,
    'Zoom_Out_As_Ctrl_Minus': 0x0049,
    'Zoom_Reset': 0x004A,
    'Zoom_Full_Screen': 0x004B,
    'Print_Screen': 0x004C,
    'Pause_Break': 0x004D,
    'Scroll_Lock': 0x004E,
    'Contextual_Menu': 0x004F,
    'Left_Button': 0x0050,  # LEFT_CLICK
    'Right_Button': 0x0051,  # RIGHT_CLICK
    'Middle_Button': 0x0052,  # MIDDLE_BUTTON
    'Back_Button': 0x0053,  # from M510v2 was BACK_AS_BUTTON_4
    'Back': 0x0054,  # BACK_AS_HID
    'Back_As_Alt_Win_Arrow': 0x0055,
    'Forward_Button': 0x0056,  # from M510v2 was FORWARD_AS_BUTTON_5
    'Forward_As_HID': 0x0057,
    'Forward_As_Alt_Win_Arrow': 0x0058,
    'Button_6': 0x0059,
    'Left_Scroll_As_Button_7': 0x005A,
    'Left_Tilt': 0x005B,  # from M510v2 was LEFT_SCROLL_AS_AC_PAN
    'Right_Scroll_As_Button_8': 0x005C,
    'Right_Tilt': 0x005D,  # from M510v2 was RIGHT_SCROLL_AS_AC_PAN
    'Button_9': 0x005E,
    'Button_10': 0x005F,
    'Button_11': 0x0060,
    'Button_12': 0x0061,
    'Button_13': 0x0062,
    'Button_14': 0x0063,
    'Button_15': 0x0064,
    'Button_16': 0x0065,
    'Button_17': 0x0066,
    'Button_18': 0x0067,
    'Button_19': 0x0068,
    'Button_20': 0x0069,
    'Button_21': 0x006A,
    'Button_22': 0x006B,
    'Button_23': 0x006C,
    'Button_24': 0x006D,
    'Show_Desktop': 0x006E,  # Craft Keyboard Fn F5
    'Lock_PC': 0x006F,  # Craft Keyboard top 1st from right
    'Fn_F1': 0x0070,
    'Fn_F2': 0x0071,
    'Fn_F3': 0x0072,
    'Fn_F4': 0x0073,
    'Fn_F5': 0x0074,
    'Fn_F6': 0x0075,
    'Fn_F7': 0x0076,
    'Fn_F8': 0x0077,
    'Fn_F9': 0x0078,
    'Fn_F10': 0x0079,
    'Fn_F11': 0x007A,
    'Fn_F12': 0x007B,
    'Fn_F13': 0x007C,
    'Fn_F14': 0x007D,
    'Fn_F15': 0x007E,
    'Fn_F16': 0x007F,
    'Fn_F17': 0x0080,
    'Fn_F18': 0x0081,
    'Fn_F19': 0x0082,
    'IOS_Home': 0x0083,
    'Android_Home': 0x0084,
    'Android_Menu': 0x0085,
    'Android_Search': 0x0086,
    'Android_Back': 0x0087,
    'Home_Combo': 0x0088,
    'Lock_Combo': 0x0089,
    'IOS_Virtual_Keyboard': 0x008A,
    'IOS_Language_Switch': 0x008B,
    'Mac_Expose': 0x008C,
    'Mac_Dashboard': 0x008D,
    'Win7_Snap_Left': 0x008E,
    'Win7_Snap_Right': 0x008F,
    'Minimize_Window': 0x0090,  # WIN7_MINIMIZE_AS_WIN_ARROW
    'Maximize_Window': 0x0091,  # WIN7_MAXIMIZE_AS_WIN_ARROW
    'Win7_Stretch_Up': 0x0092,
    'Win7_Monitor_Switch_As_Win_Shift_LeftArrow': 0x0093,
    'Win7_Monitor_Switch_As_Win_Shift_RightArrow': 0x0094,
    'Switch_Screen': 0x0095,  # WIN7_SHOW_PRESENTATION_MODE
    'Win7_Show_Mobility_Center': 0x0096,
    'Analog_HScroll': 0x0097,
    'Metro_Appswitch': 0x009F,
    'Metro_Appbar': 0x00A0,
    'Metro_Charms': 0x00A1,
    'Calc_Vkeyboard': 0x00A2,
    'Metro_Search': 0x00A3,
    'Combo_Sleep': 0x00A4,
    'Metro_Share': 0x00A5,
    'Metro_Settings': 0x00A6,
    'Metro_Devices': 0x00A7,
    'Metro_Start_Screen': 0x00A9,
    'Zoomin': 0x00AA,
    'Zoomout': 0x00AB,
    'Back_Hscroll': 0x00AC,
    'Show_Desktop_HPP': 0x00AE,
    'Fn_Left_Click': 0x00B7,  # from K400 Plus
    # https://docs.google.com/document/u/0/d/1YvXICgSe8BcBAuMr4Xu_TutvAxaa-RnGfyPFWBWzhkc/export?format=docx
    # Extract to csv.  Eliminate extra linefeeds and spaces.
    # awk -F, '/0x/{gsub(" \\+ ","_",$2); gsub("/","__",$2); gsub(" -","_Down",$2);
    # gsub(" \\+","_Up",$2); gsub("[()\"-]","",$2); gsub(" ","_",$2); printf("\t%s=0x%04X,\n", $2, $1)}' < controls.cvs
    'Second_Left_Click': 0x00B8,  # Second_LClick / on K400 Plus
    'Fn_Second_Left_Click': 0x00B9,  # Fn_Second_LClick
    'Multiplatform_App_Switch': 0x00BA,
    'Multiplatform_Home': 0x00BB,
    'Multiplatform_Menu': 0x00BC,
    'Multiplatform_Back': 0x00BD,
    'Multiplatform_Insert': 0x00BE,
    'Screen_Capture__Print_Screen': 0x00BF,  # Craft Keyboard top 3rd from right
    'Fn_Down': 0x00C0,
    'Fn_Up': 0x00C1,
    'Multiplatform_Lock': 0x00C2,
    'Mouse_Gesture_Button': 0x00C3,  # Thumb_Button on MX Master - Logitech name App_Switch_Gesture
    'Smart_Shift': 0x00C4,  # Top_Button on MX Master
    'Microphone': 0x00C5,
    'Wifi': 0x00C6,
    'Brightness_Down': 0x00C7,  # Craft Keyboard Fn F1
    'Brightness_Up': 0x00C8,  # Craft Keyboard Fn F2
    'Display_Out__Project_Screen_': 0x00C9,
    'View_Open_Apps': 0x00CA,
    'View_All_Apps': 0x00CB,
    'Switch_App': 0x00CC,
    'Fn_Inversion_Change': 0x00CD,
    'MultiPlatform_Back': 0x00CE,
    'MultiPlatform_Forward': 0x00CF,
    'MultiPlatform_Gesture_Button': 0x00D0,
    'Host_Switch_Channel_1': 0x00D1,  # Craft Keyboard
    'Host_Switch_Channel_2': 0x00D2,  # Craft Keyboard
    'Host_Switch_Channel_3': 0x00D3,  # Craft Keyboard
    'MultiPlatform_Search': 0x00D4,
    'MultiPlatform_Home__Mission_Control': 0x00D5,
    'MultiPlatform_Menu__Show__Hide_Virtual_Keyboard__Launchpad': 0x00D6,
    'Virtual_Gesture_Button': 0x00D7,
    'Cursor_Button_Long_Press': 0x00D8,
    'Next_Button_Shortpress': 0x00D9,  # Next_Button
    'Next_Button_Long_Press': 0x00DA,
    'Back_Button_Short_Press': 0x00DB,  # Back
    'Back_Button_Long_Press': 0x00DC,
    'Multi_Platform_Language_Switch': 0x00DD,
    'F_Lock': 0x00DE,
    'Switch_Highlight': 0x00DF,
    'Mission_Control__Task_View': 0x00E0,  # Craft Keyboard Fn F3 Switch_Workspace
    'Dashboard_Launchpad__Action_Center': 0x00E1,  # Craft Keyboard Fn F4 Application_Launcher
    'Backlight_Down': 0x00E2,  # Craft Keyboard Fn F6
    'Backlight_Up': 0x00E3,  # Craft Keyboard Fn F7
    'Previous_Fn': 0x00E4,  # Craft Keyboard Fn F8 Previous_Track
    'Play__Pause_Fn': 0x00E5,  # Craft Keyboard Fn F9 Play__Pause
    'Next_Fn': 0x00E6,  # Craft Keyboard Fn F10 Next_Track
    'Mute_Fn': 0x00E7,  # Craft Keyboard Fn F11 Mute
    'Volume_Down_Fn': 0x00E8,  # Craft Keyboard Fn F12 Volume_Down
    'Volume_Up_Fn': 0x00E9,  # Craft Keyboard next to F12 Volume_Down
    'App_Contextual_Menu__Right_Click': 0x00EA,  # Craft Keyboard top 2nd from right
    'Right_Arrow': 0x00EB,
    'Left_Arrow': 0x00EC,
    'DPI_Change': 0x00ED,
    'New_Tab': 0x00EE,
    'F2': 0x00EF,
    'F3': 0x00F0,
    'F4': 0x00F1,
    'F5': 0x00F2,
    'F6': 0x00F3,
    'F7': 0x00F4,
    'F8': 0x00F5,
    'F1': 0x00F6,
    'Next_Color_Effect': 0x00F7,
    'Increase_Color_Effect_Speed': 0x00F8,
    'Decrease_Color_Effect_Speed': 0x00F9,
    'Load_Lighting_Custom_Profile': 0x00FA,
    'Laser_Button_Short_Press': 0x00FB,
    'Laser_Button_Long_Press': 0x00FC,
    'DPI_Switch': 0x00FD,
    'Multiplatform_Home__Show_Desktop': 0x00FE,
    'Multiplatform_App_Switch__Show_Dashboard': 0x00FF,
    'Multiplatform_App_Switch_2': 0x0100,  # Multiplatform_App_Switch
    'Fn_Inversion__Hot_Key': 0x0101,
    'LeftAndRightClick': 0x0102,
    'Voice_Dictation': 0x0103,  # MX Keys for Business Fn F5 ; MX Mini Fn F6 Dictation
    'Emoji_Smiley_Heart_Eyes': 0x0104,
    'Emoji_Crying_Face': 0x0105,
    'Emoji_Smiley': 0x0106,
    'Emoji_Smilie_With_Tears': 0x0107,
    'Open_Emoji_Panel': 0x0108,  # MX Keys for Business Fn F6 ; MX Mini Fn F7 Emoji
    'Multiplatform_App_Switch__Launchpad': 0x0109,
    'Snipping_Tool': 0x010A,  # MX Keys for Business top 3rd from right; MX Mini Fn F8 Screenshot
    'Grave_Accent': 0x010B,
    'Tab_Key': 0x010C,
    'Caps_Lock': 0x010D,
    'Left_Shift': 0x010E,
    'Left_Control': 0x010F,
    'Left_Option__Start': 0x0110,
    'Left_Command__Alt': 0x0111,
    'Right_Command__Alt': 0x0112,
    'Right_Option__Start': 0x0113,
    'Right_Control': 0x0114,
    'Right_Shift': 0x0115,
    'Insert': 0x0116,
    'Delete': 0x0117,  # MX Mini Lock (on delete key in function row)
    'Home': 0x118,
    'End': 0x119,
    'Page_Up': 0x11A,
    'Page_Down': 0x11B,
    'Mute_Microphone': 0x11C,  # MX Keys for Business Fn F7 ; MX Mini Fn F9 Microphone Mute
    'Do_Not_Disturb': 0x11D,
    'Backslash': 0x11E,
    'Refresh': 0x11F,
    'Close_Tab': 0x120,
    'Lang_Switch': 0x121,
    'Standard_Key_A': 0x122,
    'Standard_Key_B': 0x123,
    'Standard_Key_C': 0x124,  # There are lots more of these
    'Right_Option__Start__2': 0x013C,  # On MX Mechanical Mini
    'Play_Pause': 0x0141,  # On MX Mechanical Mini
})

for i in range(1, 33):  # add in G keys - these are not really Logitech Controls
    CONTROL[0x1000 + i] = 'G' + str(i)
for i in range(1, 9):  # add in M keys - these are not really Logitech Controls
    CONTROL[0x1100 + i] = 'M' + str(i)
CONTROL[0x1200] = 'MR'  # add in MR key - this is not really a Logitech Control

CONTROL._fallback = lambda x: 'unknown:%04X' % x

# <tasks.xml awk -F\" '/<Task /{gsub(/ /, "_", $6); printf("\t%s=0x%04X,\n", $6, $4)}'
TASK = _NamedInts(
    Volume_Up=0x0001,
    Volume_Down=0x0002,
    Mute=0x0003,
    # Multimedia tasks:
    Play__Pause=0x0004,
    Next=0x0005,
    Previous=0x0006,
    Stop=0x0007,
    Application_Switcher=0x0008,
    BurnMediaPlayer=0x0009,
    Calculator=0x000A,
    Calendar=0x000B,
    Close_Application=0x000C,
    Eject=0x000D,
    Email=0x000E,
    Help=0x000F,
    OffDocument=0x0010,
    OffSpreadsheet=0x0011,
    OffPowerpnt=0x0012,
    Undo=0x0013,
    Redo=0x0014,
    Print=0x0015,
    Save=0x0016,
    SmartKeySet=0x0017,
    Favorites=0x0018,
    GadgetsSet=0x0019,
    HomePage=0x001A,
    WindowsRestore=0x001B,
    WindowsMinimize=0x001C,
    Music=0x001D,  # also known as MediaPlayer
    # Both 0x001E and 0x001F are known as MediaCenterSet
    Media_Center_Logitech=0x001E,
    Media_Center_Microsoft=0x001F,
    UserMenu=0x0020,
    Messenger=0x0021,
    PersonalFolders=0x0022,
    MyMusic=0x0023,
    Webcam=0x0024,
    PicturesFolder=0x0025,
    MyVideos=0x0026,
    My_Computer=0x0027,
    PictureAppSet=0x0028,
    Search=0x0029,  # also known as AdvSmartSearch
    RecordMediaPlayer=0x002A,
    BrowserRefresh=0x002B,
    RotateRight=0x002C,
    Search_Files=0x002D,  # SearchForFiles
    MM_SHUFFLE=0x002E,
    Sleep=0x002F,  # also known as StandBySet
    BrowserStop=0x0030,
    OneTouchSync=0x0031,
    ZoomSet=0x0032,
    ZoomBtnInSet2=0x0033,
    ZoomBtnInSet=0x0034,
    ZoomBtnOutSet2=0x0035,
    ZoomBtnOutSet=0x0036,
    ZoomBtnResetSet=0x0037,
    Left_Click=0x0038,  # LeftClick
    Right_Click=0x0039,  # RightClick
    Mouse_Middle_Button=0x003A,  # from M510v2 was MiddleMouseButton
    Back=0x003B,
    Mouse_Back_Button=0x003C,  # from M510v2 was BackEx
    BrowserForward=0x003D,
    Mouse_Forward_Button=0x003E,  # from M510v2 was BrowserForwardEx
    Mouse_Scroll_Left_Button_=0x003F,  # from M510v2 was HorzScrollLeftSet
    Mouse_Scroll_Right_Button=0x0040,  # from M510v2 was HorzScrollRightSet
    QuickSwitch=0x0041,
    BatteryStatus=0x0042,
    Show_Desktop=0x0043,  # ShowDesktop
    WindowsLock=0x0044,
    FileLauncher=0x0045,
    FolderLauncher=0x0046,
    GotoWebAddress=0x0047,
    GenericMouseButton=0x0048,
    KeystrokeAssignment=0x0049,
    LaunchProgram=0x004A,
    MinMaxWindow=0x004B,
    VOLUMEMUTE_NoOSD=0x004C,
    New=0x004D,
    Copy=0x004E,
    CruiseDown=0x004F,
    CruiseUp=0x0050,
    Cut=0x0051,
    Do_Nothing=0x0052,
    PageDown=0x0053,
    PageUp=0x0054,
    Paste=0x0055,
    SearchPicture=0x0056,
    Reply=0x0057,
    PhotoGallerySet=0x0058,
    MM_REWIND=0x0059,
    MM_FASTFORWARD=0x005A,
    Send=0x005B,
    ControlPanel=0x005C,
    UniversalScroll=0x005D,
    AutoScroll=0x005E,
    GenericButton=0x005F,
    MM_NEXT=0x0060,
    MM_PREVIOUS=0x0061,
    Do_Nothing_One=0x0062,  # also known as Do_Nothing
    SnapLeft=0x0063,
    SnapRight=0x0064,
    WinMinRestore=0x0065,
    WinMaxRestore=0x0066,
    WinStretch=0x0067,
    SwitchMonitorLeft=0x0068,
    SwitchMonitorRight=0x0069,
    ShowPresentation=0x006A,
    ShowMobilityCenter=0x006B,
    HorzScrollNoRepeatSet=0x006C,
    TouchBackForwardHorzScroll=0x0077,
    MetroAppSwitch=0x0078,
    MetroAppBar=0x0079,
    MetroCharms=0x007A,
    Calculator_VKEY=0x007B,  # also known as Calculator
    MetroSearch=0x007C,
    MetroStartScreen=0x0080,
    MetroShare=0x007D,
    MetroSettings=0x007E,
    MetroDevices=0x007F,
    MetroBackLeftHorz=0x0082,
    MetroForwRightHorz=0x0083,
    Win8_Back=0x0084,  # also known as MetroCharms
    Win8_Forward=0x0085,  # also known as AppSwitchBar
    Win8Charm_Appswitch_GifAnimation=0x0086,
    Win8BackHorzLeft=0x008B,  # also known as Back
    Win8ForwardHorzRight=0x008C,  # also known as BrowserForward
    MetroSearch2=0x0087,
    MetroShare2=0x0088,
    MetroSettings2=0x008A,
    MetroDevices2=0x0089,
    Win8MetroWin7Forward=0x008D,  # also known as MetroStartScreen
    Win8ShowDesktopWin7Back=0x008E,  # also known as ShowDesktop
    MetroApplicationSwitch=0x0090,  # also known as MetroStartScreen
    ShowUI=0x0092,
    # https://docs.google.com/document/d/1Dpx_nWRQAZox_zpZ8SNc9nOkSDE9svjkghOCbzopabc/edit
    # Extract to csv.  Eliminate extra linefeeds and spaces. Turn / into __ and space into _
    # awk -F, '/0x/{gsub(" \\+ ","_",$2);  gsub("_-","_Down",$2); gsub("_\\+","_Up",$2);
    # gsub("[()\"-]","",$2); gsub(" ","_",$2); printf("\t%s=0x%04X,\n", $2, $1)}' < tasks.csv > tasks.py
    Switch_Presentation__Switch_Screen=0x0093,  # on K400 Plus
    Minimize_Window=0x0094,
    Maximize_Window=0x0095,  # on K400 Plus
    MultiPlatform_App_Switch=0x0096,
    MultiPlatform_Home=0x0097,
    MultiPlatform_Menu=0x0098,
    MultiPlatform_Back=0x0099,
    Switch_Language=0x009A,  # Mac_switch_language
    Screen_Capture=0x009B,  # Mac_screen_Capture, on Craft Keyboard
    Gesture_Button=0x009C,
    Smart_Shift=0x009D,
    AppExpose=0x009E,
    Smart_Zoom=0x009F,
    Lookup=0x00A0,
    Microphone_on__off=0x00A1,
    Wifi_on__off=0x00A2,
    Brightness_Down=0x00A3,
    Brightness_Up=0x00A4,
    Display_Out=0x00A5,
    View_Open_Apps=0x00A6,
    View_All_Open_Apps=0x00A7,
    AppSwitch=0x00A8,
    Gesture_Button_Navigation=0x00A9,  # Mouse_Thumb_Button on MX Master
    Fn_inversion=0x00AA,
    Multiplatform_Back=0x00AB,
    Multiplatform_Forward=0x00AC,
    Multiplatform_Gesture_Button=0x00AD,
    HostSwitch_Channel_1=0x00AE,
    HostSwitch_Channel_2=0x00AF,
    HostSwitch_Channel_3=0x00B0,
    Multiplatform_Search=0x00B1,
    Multiplatform_Home__Mission_Control=0x00B2,
    Multiplatform_Menu__Launchpad=0x00B3,
    Virtual_Gesture_Button=0x00B4,
    Cursor=0x00B5,
    Keyboard_Right_Arrow=0x00B6,
    SW_Custom_Highlight=0x00B7,
    Keyboard_Left_Arrow=0x00B8,
    TBD=0x00B9,
    Multiplatform_Language_Switch=0x00BA,
    SW_Custom_Highlight_2=0x00BB,
    Fast_Forward=0x00BC,
    Fast_Backward=0x00BD,
    Switch_Highlighting=0x00BE,
    Mission_Control__Task_View=0x00BF,  # Switch_Workspace on Craft Keyboard
    Dashboard_Launchpad__Action_Center=0x00C0,  # Application_Launcher on Craft Keyboard
    Backlight_Down=0x00C1,  # Backlight_Down_FW_internal_function
    Backlight_Up=0x00C2,  # Backlight_Up_FW_internal_function
    Right_Click__App_Contextual_Menu=0x00C3,  # Context_Menu on Craft Keyboard
    DPI_Change=0x00C4,
    New_Tab=0x00C5,
    F2=0x00C6,
    F3=0x00C7,
    F4=0x00C8,
    F5=0x00C9,
    F6=0x00CA,
    F7=0x00CB,
    F8=0x00CC,
    F1=0x00CD,
    Laser_Button=0x00CE,
    Laser_Button_Long_Press=0x00CF,
    Start_Presentation=0x00D0,
    Blank_Screen=0x00D1,
    DPI_Switch=0x00D2,  # AdjustDPI on MX Vertical
    Home__Show_Desktop=0x00D3,
    App_Switch__Dashboard=0x00D4,
    App_Switch=0x00D5,
    Fn_Inversion=0x00D6,
    LeftAndRightClick=0x00D7,
    Voice_Dictation=0x00D8,
    Emoji_Smiling_Face_With_Heart_Shaped_Eyes=0x00D9,
    Emoji_Loudly_Crying_Face=0x00DA,
    Emoji_Smiley=0x00DB,
    Emoji_Smiley_With_Tears=0x00DC,
    Open_Emoji_Panel=0x00DD,
    Multiplatform_App_Switch__Launchpad=0x00DE,
    Snipping_Tool=0x00DF,
    Grave_Accent=0x00E0,
    Standard_Tab_Key=0x00E1,
    Caps_Lock=0x00E2,
    Left_Shift=0x00E3,
    Left_Control=0x00E4,
    Left_Option__Start=0x00E5,
    Left_Command__Alt=0x00E6,
    Right_Command__Alt=0x00E7,
    Right_Option__Start=0x00E8,
    Right_Control=0x00E9,
    Right_Shift=0x0EA,
    Insert=0x00EB,
    Delete=0x00EC,
    Home=0x00ED,
    End=0x00EE,
    Page_Up=0x00EF,
    Page_Down=0x00F0,
    Mute_Microphone=0x00F1,
    Do_Not_Disturb=0x00F2,
    Backslash=0x00F3,
    Refresh=0x00F4,
    Close_Tab=0x00F5,
    Lang_Switch=0x00F6,
    Standard_Alphabetical_Key=0x00F7,
    Right_Option__Start__2=0x00F8,
    Left_Option=0x00F9,
    Right_Option=0x00FA,
    Left_Cmd=0x00FB,
    Right_Cmd=0x00FC,
)
TASK._fallback = lambda x: 'unknown:%04X' % x
# Capabilities and desired software handling for a control
# Ref: https://drive.google.com/file/d/10imcbmoxTJ1N510poGdsviEhoFfB_Ua4/view
# We treat bytes 4 and 8 of `getCidInfo` as a single bitfield
KEY_FLAG = _NamedInts(
    analytics_key_events=0x400,
    force_raw_XY=0x200,
    raw_XY=0x100,
    virtual=0x80,
    persistently_divertable=0x40,
    divertable=0x20,
    reprogrammable=0x10,
    FN_sensitive=0x08,
    nonstandard=0x04,
    is_FN=0x02,
    mse=0x01
)
# Flags describing the reporting method of a control
# We treat bytes 2 and 5 of `get/setCidReporting` as a single bitfield
MAPPING_FLAG = _NamedInts(
    analytics_key_events_reporting=0x100,
    force_raw_XY_diverted=0x40,
    raw_XY_diverted=0x10,
    persistently_diverted=0x04,
    diverted=0x01
)
CID_GROUP_BIT = _NamedInts(g8=0x80, g7=0x40, g6=0x20, g5=0x10, g4=0x08, g3=0x04, g2=0x02, g1=0x01)
CID_GROUP = _NamedInts(g8=8, g7=7, g6=6, g5=5, g4=4, g3=3, g2=2, g1=1)
DISABLE = _NamedInts(
    Caps_Lock=0x01,
    Num_Lock=0x02,
    Scroll_Lock=0x04,
    Insert=0x08,
    Win=0x10,  # aka Super
)
DISABLE._fallback = lambda x: 'unknown:%02X' % x

# HID USB Keycodes from https://www.usb.org/sites/default/files/documents/hut1_12v2.pdf
# Modified by information from Linux HID driver linux/drivers/hid/hid-input.c
USB_HID_KEYCODES = _NamedInts(
    A=0x04,
    B=0x05,
    C=0x06,
    D=0x07,
    E=0x08,
    F=0x09,
    G=0x0a,
    H=0x0b,
    I=0x0c,
    J=0x0d,
    K=0x0e,
    L=0x0f,
    M=0x10,
    N=0x11,
    O=0x12,
    P=0x13,
    Q=0x14,
    R=0x15,
    S=0x16,
    T=0x17,
    U=0x18,
    V=0x19,
    W=0x1a,
    X=0x1b,
    Y=0x1c,
    Z=0x1d,
    ENTER=0x28,
    ESC=0x29,
    BACKSPACE=0x2a,
    TAB=0x2b,
    SPACE=0x2c,
    MINUS=0x2d,
    EQUAL=0x2e,
    LEFTBRACE=0x2f,
    RIGHTBRACE=0x30,
    BACKSLASH=0x31,
    HASHTILDE=0x32,
    SEMICOLON=0x33,
    APOSTROPHE=0x34,
    GRAVE=0x35,
    COMMA=0x36,
    DOT=0x37,
    SLASH=0x38,
    CAPSLOCK=0x39,
    F1=0x3a,
    F2=0x3b,
    F3=0x3c,
    F4=0x3d,
    F5=0x3e,
    F6=0x3f,
    F7=0x40,
    F8=0x41,
    F9=0x42,
    F10=0x43,
    F11=0x44,
    F12=0x45,
    SYSRQ=0x46,
    SCROLLLOCK=0x47,
    PAUSE=0x48,
    INSERT=0x49,
    HOME=0x4a,
    PAGEUP=0x4b,
    DELETE=0x4c,
    END=0x4d,
    PAGEDOWN=0x4e,
    RIGHT=0x4f,
    LEFT=0x50,
    DOWN=0x51,
    UP=0x52,
    NUMLOCK=0x53,
    KPSLASH=0x54,
    KPASTERISK=0x55,
    KPMINUS=0x56,
    KPPLUS=0x57,
    KPENTER=0x58,
    KP1=0x59,
    KP2=0x5a,
    KP3=0x5b,
    KP4=0x5c,
    KP5=0x5d,
    KP6=0x5e,
    KP7=0x5f,
    KP8=0x60,
    KP9=0x61,
    KP0=0x62,
    KPDOT=0x63,
    COMPOSE=0x65,
    POWER=0x66,
    KPEQUAL=0x67,
    F13=0x68,
    F14=0x69,
    F15=0x6a,
    F16=0x6b,
    F17=0x6c,
    F18=0x6d,
    F19=0x6e,
    F20=0x6f,
    F21=0x70,
    F22=0x71,
    F23=0x72,
    F24=0x73,
    OPEN=0x74,
    HELP=0x75,
    PROPS=0x76,
    FRONT=0x77,
    STOP=0x78,
    AGAIN=0x79,
    UNDO=0x7a,
    CUT=0x7b,
    COPY=0x7c,
    PASTE=0x7d,
    FIND=0x7e,
    MUTE=0x7f,
    VOLUMEUP=0x80,
    VOLUMEDOWN=0x81,
    KPCOMMA=0x85,
    RO=0x87,
    KATAKANAHIRAGANA=0x88,
    YEN=0x89,
    HENKAN=0x8a,
    MUHENKAN=0x8b,
    KPJPCOMMA=0x8c,
    HANGEUL=0x90,
    HANJA=0x91,
    KATAKANA=0x92,
    HIRAGANA=0x93,
    ZENKAKUHANKAKU=0x94,
    KPLEFTPAREN=0xb6,
    KPRIGHTPAREN=0xb7,
    LEFTCTRL=0xe0,
    LEFTSHIFT=0xe1,
    LEFTALT=0xe2,
    LEFTWINDOWS=0xe3,
    RIGHTCTRL=0xe4,
    RIGHTSHIFT=0xe5,
    RIGHTALT=0xe6,
    RIGHTMETA=0xe7,
    MEDIA_PLAYPAUSE=0xe8,
    MEDIA_STOPCD=0xe9,
    MEDIA_PREVIOUSSONG=0xea,
    MEDIA_NEXTSONG=0xeb,
    MEDIA_EJECTCD=0xec,
    MEDIA_VOLUMEUP=0xed,
    MEDIA_VOLUMEDOWN=0xee,
    MEDIA_MUTE=0xef,
    MEDIA_WWW=0xf0,
    MEDIA_BACK=0xf1,
    MEDIA_FORWARD=0xf2,
    MEDIA_STOP=0xf3,
    MEDIA_FIND=0xf4,
    MEDIA_SCROLLUP=0xf5,
    MEDIA_SCROLLDOWN=0xf6,
    MEDIA_EDIT=0xf7,
    MEDIA_SLEEP=0xf8,
    MEDIA_COFFEE=0xf9,
    MEDIA_REFRESH=0xfa,
    MEDIA_CALC=0xfb,
)
USB_HID_KEYCODES[0] = 'No Output'
USB_HID_KEYCODES[0x1e] = '1'
USB_HID_KEYCODES[0x1f] = '2'
USB_HID_KEYCODES[0x20] = '3'
USB_HID_KEYCODES[0x21] = '4'
USB_HID_KEYCODES[0x22] = '5'
USB_HID_KEYCODES[0x23] = '6'
USB_HID_KEYCODES[0x24] = '7'
USB_HID_KEYCODES[0x25] = '8'
USB_HID_KEYCODES[0x26] = '9'
USB_HID_KEYCODES[0x27] = '0'
USB_HID_KEYCODES[0x64] = '102ND'

HID_CONSUMERCODES = _NamedInts({
    #    Unassigned=0x00,
    #    Consumer_Control=0x01,
    #    Numeric_Key_Pad=0x02,
    #    Programmable_Buttons=0x03,
    #    Microphone=0x04,
    #    Headphone=0x05,
    #    Graphic_Equalizer=0x06,
    #    AM__PM=0x22,
    'Power': 0x30,
    'Reset': 0x31,
    'Sleep': 0x32,
    'Sleep_After': 0x33,
    'Sleep_Mode': 0x34,
    'Illumination': 0x35,
    'Function_Buttons': 0x36,
    'Menu': 0x40,
    'Menu__Pick': 0x41,
    'Menu_Up': 0x42,
    'Menu_Down': 0x43,
    'Menu_Left': 0x44,
    'Menu_Right': 0x45,
    'Menu_Escape': 0x46,
    'Menu_Value_Increase': 0x47,
    'Menu_Value_Decrease': 0x48,
    'Data_On_Screen': 0x60,
    'Closed_Caption': 0x61,
    #    Closed_Caption_Select=0x62,
    'VCR__TV': 0x63,
    #    Broadcast_Mode=0x64,
    'Snapshot': 0x65,
    #    Still=0x66,
    'Red': 0x69,
    'Green': 0x6A,
    'Blue': 0x6B,
    'Yellow': 0x6C,
    'Aspect_Ratio': 0x6D,
    'Brightness_Up': 0x6F,
    'Brightness_Down': 0x70,
    'Brightness_Toggle': 0x72,
    'Brightness_Min': 0x73,
    'Brightness_Max': 0x74,
    'Brightness_Auto': 0x75,
    'Keyboard_Illumination_Up': 0x79,
    'Keyboard_Illumination_Down': 0x7A,
    'Keyboard_Illumination_Toggle': 0x7C,
    #    Selection=0x80,
    #    Assign_Selection=0x81,
    'Mode_Step': 0x82,
    'Recall_Last': 0x83,
    'Enter_Channel': 0x84,
    #    Order_Movie=0x85,
    #    Channel=0x86,
    #    Media_Selection=0x87,
    'Media_Select_Computer': 0x88,
    'Media_Select_TV': 0x89,
    'Media_Select_WWW': 0x8A,
    'Media_Select_DVD': 0x8B,
    'Media_Select_Telephone': 0x8C,
    'Media_Select_Program_Guide': 0x8D,
    'Media_Select_Video_Phone': 0x8E,
    'Media_Select_Games': 0x8F,
    'Media_Select_Messages': 0x90,
    'Media_Select_CD': 0x91,
    'Media_Select_VCR': 0x92,
    'Media_Select_Tuner': 0x93,
    'Quit': 0x94,
    'Help': 0x95,
    'Media_Select_Tape': 0x96,
    'Media_Select_Cable': 0x97,
    'Media_Select_Satellite': 0x98,
    'Media_Select_Security': 0x99,
    'Media_Select_Home': 0x9A,
    #    Media_Select_Call=0x9B,
    'Channel_Increment': 0x9C,
    'Channel_Decrement': 0x9D,
    #    Media_Select_SAP=0x9E,
    'VCR_Plus': 0xA0,
    #    Once=0xA1,
    #    Daily=0xA2,
    #    Weekly=0xA3,
    #    Monthly=0xA4,
    'Play': 0xB0,
    'Pause': 0xB1,
    'Record': 0xB2,
    'Fast_Forward': 0xB3,
    'Rewind': 0xB4,
    'Scan_Next_Track': 0xB5,
    'Scan_Previous_Track': 0xB6,
    'Stop': 0xB7,
    'Eject': 0xB8,
    'Random_Play': 0xB9,
    'Select_DisC': 0xBA,
    'Enter_Disc': 0xBB,
    'Repeat': 0xBC,
    'Tracking': 0xBD,
    'Track_Normal': 0xBE,
    'Slow_Tracking': 0xBF,
    #    Frame_Forward=0xC0,
    #    Frame_Back=0xC1,
    #    Mark=0xC2,
    #    Clear_Mark=0xC3,
    #    Repeat_From_Mark=0xC4,
    #    Return_To_Mark=0xC5,
    #    Search_Mark_Forward=0xC6,
    #    Search_Mark_Backwards=0xC7,
    #    Counter_Reset=0xC8,
    #    Show_Counter=0xC9,
    #    Tracking_Increment=0xCA,
    #    Tracking_Decrement=0xCB,
    #    Stop__Eject=0xCC,
    'Play__Pause': 0xCD,
    #    Play__Skip=0xCE,
    'Volume': 0xE0,
    #    Balance=0xE1,
    'Mute': 0xE2,
    #    Bass=0xE3,
    #    Treble=0xE4,
    'Bass_Boost': 0xE5,
    #    Surround_Mode=0xE6,
    #    Loudness=0xE7,
    #    MPX=0xE8,
    'Volume_Up': 0xE9,
    'Volume_Down': 0xEA,
    #    Speed_Select=0xF0,
    #    Playback_Speed=0xF1,
    #    Standard_Play=0xF2,
    #    Long_Play=0xF3,
    #    Extended_Play=0xF4,
    'Slow': 0xF5,
    'Fan_Enable': 0x100,
    'Fan_Speed': 0x101,
    'Light': 0x102,
    'Light_Illumination_Level': 0x103,
    'Climate_Control_Enable': 0x104,
    'Room_Temperature': 0x105,
    'Security_Enable': 0x106,
    'Fire_Alarm': 0x107,
    'Police_Alarm': 0x108,
    'Proximity': 0x109,
    'Motion': 0x10A,
    'Duress_Alarm': 0x10B,
    'Holdup_Alarm': 0x10C,
    'Medical_Alarm': 0x10D,
    'Balance_Right': 0x150,
    'Balance_Left': 0x151,
    'Bass_Increment': 0x152,
    'Bass_Decrement': 0x153,
    'Treble_Increment': 0x154,
    'Treble_Decrement': 0x155,
    'Speaker_System': 0x160,
    'Channel_Left': 0x161,
    'Channel_Right': 0x162,
    'Channel_Center': 0x163,
    'Channel_Front': 0x164,
    'Channel_Center_Front': 0x165,
    'Channel_Side': 0x166,
    'Channel_Surround': 0x167,
    'Channel_Low_Frequency_Enhancement': 0x168,
    'Channel_Top': 0x169,
    'Channel_Unknown': 0x16A,
    'Subchannel': 0x170,
    'Subchannel_Increment': 0x171,
    'Subchannel_Decrement': 0x172,
    'Alternate_Audio_Increment': 0x173,
    'Alternate_Audio_Decrement': 0x174,
    'Application_Launch_Buttons': 0x180,
    'AL_Launch_Button_Configuration_Tool': 0x181,
    'AL_Programmable_Button_Configuration': 0x182,
    'AL_Consumer_Control_Configuration': 0x183,
    'AL_Word_Processor': 0x184,
    'AL_Text_Editor': 0x185,
    'AL_Spreadsheet': 0x186,
    'AL_Graphics_Editor': 0x187,
    'AL_Presentation_App': 0x188,
    'AL_Database_App': 0x189,
    'AL_Email_Reader': 0x18A,
    'AL_Newsreader': 0x18B,
    'AL_Voicemail': 0x18C,
    'AL_Contacts__Address_Book': 0x18D,
    'AL_Calendar__Schedule': 0x18E,
    'AL_Task__Project_Manager': 0x18F,
    'AL_Log__Journal__Timecard': 0x190,
    'AL_Checkbook__Finance': 0x191,
    'AL_Calculator': 0x192,
    'AL_A__V_Capture__Playback': 0x193,
    'AL_Local_Machine_Browser': 0x194,
    'AL_LAN__WAN_Browser': 0x195,
    'AL_Internet_Browser': 0x196,
    'AL_Remote_Networking__ISP_Connect': 0x197,
    'AL_Network_Conference': 0x198,
    'AL_Network_Chat': 0x199,
    'AL_Telephony__Dialer': 0x19A,
    'AL_Logon': 0x19B,
    'AL_Logoff': 0x19C,
    'AL_Logon__Logoff': 0x19D,
    'AL_Terminal_Lock__Screensaver': 0x19E,
    'AL_Control_Panel': 0x19F,
    'AL_Command_Line_Processor__Run': 0x1A0,
    'AL_Process__Task_Manager': 0x1A1,
    'AL_Select_Tast__Application': 0x1A2,
    'AL_Next_Task__Application': 0x1A3,
    'AL_Previous_Task__Application': 0x1A4,
    'AL_Preemptive_Halt_Task__Application': 0x1A5,
    'AL_Integrated_Help_Center': 0x1A6,
    'AL_Documents': 0x1A7,
    'AL_Thesaurus': 0x1A8,
    'AL_Dictionary': 0x1A9,
    'AL_Desktop': 0x1AA,
    'AL_Spell_Check': 0x1AB,
    'AL_Grammar_Check': 0x1AC,
    'AL_Wireless_Status': 0x1AD,
    'AL_Keyboard_Layout': 0x1AE,
    'AL_Virus_Protection': 0x1AF,
    'AL_Encryption': 0x1B0,
    'AL_Screen_Saver': 0x1B1,
    'AL_Alarms': 0x1B2,
    'AL_Clock': 0x1B3,
    'AL_File_Browser': 0x1B4,
    'AL_Power_Status': 0x1B5,
    'AL_Image_Browser': 0x1B6,
    'AL_Audio_Browser': 0x1B7,
    'AL_Movie_Browser': 0x1B8,
    'AL_Digital_Rights_Manager': 0x1B9,
    'AL_Digital_Wallet': 0x1BA,
    'AL_Instant_Messaging': 0x1BC,
    'AL_OEM_Features___Tips__Tutorial_Browser': 0x1BD,
    'AL_OEM_Help': 0x1BE,
    'AL_Online_Community': 0x1BF,
    'AL_Entertainment_Content_Browser': 0x1C0,
    'AL_Online_Shopping_Browser': 0x1C1,
    'AL_SmartCard_Information__Help': 0x1C2,
    'AL_Market_Monitor__Finance_Browser': 0x1C3,
    'AL_Customized_Corporate_News_Browser': 0x1C4,
    'AL_Online_Activity_Browser': 0x1C5,
    'AL_Research__Search_Browser': 0x1C6,
    'AL_Audio_Player': 0x1C7,
    'Generic_GUI_Application_Controls': 0x200,
    'AC_New': 0x201,
    'AC_Open': 0x202,
    'AC_Close': 0x203,
    'AC_Exit': 0x204,
    'AC_Maximize': 0x205,
    'AC_Minimize': 0x206,
    'AC_Save': 0x207,
    'AC_Print': 0x208,
    'AC_Properties': 0x209,
    'AC_Undo': 0x21A,
    'AC_Copy': 0x21B,
    'AC_Cut': 0x21C,
    'AC_Paste': 0x21D,
    'AC_Select_All': 0x21E,
    'AC_Find': 0x21F,
    'AC_Find_and_Replace': 0x220,
    'AC_Search': 0x221,
    'AC_Go_To': 0x222,
    'AC_Home': 0x223,
    'AC_Back': 0x224,
    'AC_Forward': 0x225,
    'AC_Stop': 0x226,
    'AC_Refresh': 0x227,
    'AC_Previous_Link': 0x228,
    'AC_Next_Link': 0x229,
    'AC_Bookmarks': 0x22A,
    'AC_History': 0x22B,
    'AC_Subscriptions': 0x22C,
    'AC_Zoom_In': 0x22D,
    'AC_Zoom_Out': 0x22E,
    'AC_Zoom': 0x22F,
    'AC_Full_Screen_View': 0x230,
    'AC_Normal_View': 0x231,
    'AC_View_Toggle': 0x232,
    'AC_Scroll_Up': 0x233,
    'AC_Scroll_Down': 0x234,
    'AC_Scroll': 0x235,
    'AC_Pan_Left': 0x236,
    'AC_Pan_Right': 0x237,
    'AC_Pan': 0x238,
    'AC_New_Window': 0x239,
    'AC_Tile_Horizontally': 0x23A,
    'AC_Tile_Vertically': 0x23B,
    'AC_Format': 0x23C,
    'AC_Edit': 0x23D,
    'AC_Bold': 0x23E,
    'AC_Italics': 0x23F,
    'AC_Underline': 0x240,
    'AC_Strikethrough': 0x241,
    'AC_Subscript': 0x242,
    'AC_Superscript': 0x243,
    'AC_All_Caps': 0x244,
    'AC_Rotate': 0x245,
    'AC_Resize': 0x246,
    'AC_Flip_horizontal': 0x247,
    'AC_Flip_Vertical': 0x248,
    'AC_Mirror_Horizontal': 0x249,
    'AC_Mirror_Vertical': 0x24A,
    'AC_Font_Select': 0x24B,
    'AC_Font_Color': 0x24C,
    'AC_Font_Size': 0x24D,
    'AC_Justify_Left': 0x24E,
    'AC_Justify_Center_H': 0x24F,
    'AC_Justify_Right': 0x250,
    'AC_Justify_Block_H': 0x251,
    'AC_Justify_Top': 0x252,
    'AC_Justify_Center_V': 0x253,
    'AC_Justify_Bottom': 0x254,
    'AC_Justify_Block_V': 0x255,
    'AC_Indent_Decrease': 0x256,
    'AC_Indent_Increase': 0x257,
    'AC_Numbered_List': 0x258,
    'AC_Restart_Numbering': 0x259,
    'AC_Bulleted_List': 0x25A,
    'AC_Promote': 0x25B,
    'AC_Demote': 0x25C,
    'AC_Yes': 0x25D,
    'AC_No': 0x25E,
    'AC_Cancel': 0x25F,
    'AC_Catalog': 0x260,
    'AC_Buy__Checkout': 0x261,
    'AC_Add_to_Cart': 0x262,
    'AC_Expand': 0x263,
    'AC_Expand_All': 0x264,
    'AC_Collapse': 0x265,
    'AC_Collapse_All': 0x266,
    'AC_Print_Preview': 0x267,
    'AC_Paste_Special': 0x268,
    'AC_Insert_Mode': 0x269,
    'AC_Delete': 0x26A,
    'AC_Lock': 0x26B,
    'AC_Unlock': 0x26C,
    'AC_Protect': 0x26D,
    'AC_Unprotect': 0x26E,
    'AC_Attach_Comment': 0x26F,
    'AC_Delete_Comment': 0x270,
    'AC_View_Comment': 0x271,
    'AC_Select_Word': 0x272,
    'AC_Select_Sentence': 0x273,
    'AC_Select_Paragraph': 0x274,
    'AC_Select_Column': 0x275,
    'AC_Select_Row': 0x276,
    'AC_Select_Table': 0x277,
    'AC_Select_Object': 0x278,
    'AC_Redo__Repeat': 0x279,
    'AC_Sort': 0x27A,
    'AC_Sort_Ascending': 0x27B,
    'AC_Sort_Descending': 0x27C,
    'AC_Filter': 0x27D,
    'AC_Set_Clock': 0x27E,
    'AC_View_Clock': 0x27F,
    'AC_Select_Time_Zone': 0x280,
    'AC_Edit_Time_Zones': 0x281,
    'AC_Set_Alarm': 0x282,
    'AC_Clear_Alarm': 0x283,
    'AC_Snooze_Alarm': 0x284,
    'AC_Reset_Alarm': 0x285,
    'AC_Synchronize': 0x286,
    'AC_Send__Receive': 0x287,
    'AC_Send_To': 0x288,
    'AC_Reply': 0x289,
    'AC_Reply_All': 0x28A,
    'AC_Forward_Msg': 0x28B,
    'AC_Send': 0x28C,
    'AC_Attach_File': 0x28D,
    'AC_Upload': 0x28E,
    'AC_Download_Save_Target_As': 0x28F,
    'AC_Set_Borders': 0x290,
    'AC_Insert_Row': 0x291,
    'AC_Insert_Column': 0x292,
    'AC_Insert_File': 0x293,
    'AC_Insert_Picture': 0x294,
    'AC_Insert_Object': 0x295,
    'AC_Insert_Symbol': 0x296,
    'AC_Save_and_Close': 0x297,
    'AC_Rename': 0x298,
    'AC_Merge': 0x299,
    'AC_Split': 0x29A,
    'AC_Distribute_Horizontally': 0x29B,
    'AC_Distribute_Vertically': 0x29C,
})
HID_CONSUMERCODES[0x20] = '+10'
HID_CONSUMERCODES[0x21] = '+100'
HID_CONSUMERCODES._fallback = lambda x: 'unknown:%04X' % x

## Information for x1c00 Persistent from https://drive.google.com/drive/folders/0BxbRzx7vEV7eWmgwazJ3NUFfQ28

KEYMOD = _NamedInts(CTRL=0x01, SHIFT=0x02, ALT=0x04, META=0x08, RCTRL=0x10, RSHIFT=0x20, RALT=0x40, RMETA=0x80)

ACTIONID = _NamedInts(
    Empty=0x00,
    Key=0x01,
    Mouse=0x02,
    Xdisp=0x03,
    Ydisp=0x04,
    Vscroll=0x05,
    Hscroll=0x06,
    Consumer=0x07,
    Internal=0x08,
    Power=0x09
)

MOUSE_BUTTONS = _NamedInts(
    Mouse_Button_Left=0x0001,
    Mouse_Button_Right=0x0002,
    Mouse_Button_Middle=0x0004,
    Mouse_Button_Back=0x0008,
    Mouse_Button_Forward=0x0010,
    Mouse_Button_6=0x0020,
    Mouse_Button_Scroll_Left=0x0040,
    Mouse_Button_Scroll_Right=0x0080,
    Mouse_Button_9=0x0100,
    Mouse_Button_10=0x0200,
    Mouse_Button_11=0x0400,
    Mouse_Button_12=0x0800,
    Mouse_Button_13=0x1000,
    Mouse_Button_DPI=0x2000,
    Mouse_Button_15=0x4000,
    Mouse_Button_16=0x8000,
)
MOUSE_BUTTONS._fallback = lambda x: 'unknown mouse button:%04X' % x

HORIZONTAL_SCROLL = _NamedInts(
    Horizontal_Scroll_Left=0x4000,
    Horizontal_Scroll_Right=0x8000,
)
HORIZONTAL_SCROLL._fallback = lambda x: 'unknown horizontal scroll:%04X' % x

# Construct universe for Persistent Remappable Keys setting (only for supported values)
KEYS = _UnsortedNamedInts()
KEYS_Default = 0x7FFFFFFF  # Special value to reset key to default - has to be different from all others
KEYS[KEYS_Default] = 'Default'  # Value to reset to default
KEYS[0] = 'None'  # Value for no output

# Add HID keys plus modifiers
modifiers = {
    0x00: '',
    0x01: 'Cntrl+',
    0x02: 'Shift+',
    0x04: 'Alt+',
    0x08: 'Meta+',
    0x03: 'Cntrl+Shift+',
    0x05: 'Alt+Cntrl+',
    0x09: 'Meta+Cntrl+',
    0x06: 'Alt+Shift+',
    0x0A: 'Meta+Shift+',
    0x0C: 'Meta+Alt+'
}
for val, name in modifiers.items():
    for key in USB_HID_KEYCODES:
        KEYS[(ACTIONID.Key << 24) + (int(key) << 8) + val] = name + str(key)

# Add HID Consumer Codes
for code in HID_CONSUMERCODES:
    KEYS[(ACTIONID.Consumer << 24) + (int(code) << 8)] = str(code)

# Add Mouse Buttons
for code in MOUSE_BUTTONS:
    KEYS[(ACTIONID.Mouse << 24) + (int(code) << 8)] = str(code)

# Add Horizontal Scroll
for code in HORIZONTAL_SCROLL:
    KEYS[(ACTIONID.Hscroll << 24) + (int(code) << 8)] = str(code)


# Construct subsets for known devices
def persistent_keys(action_ids):
    keys = _UnsortedNamedInts()
    keys[KEYS_Default] = 'Default'  # Value to reset to default
    keys[0] = 'No Output (only as default)'
    for key in KEYS:
        if (int(key) >> 24) in action_ids:
            keys[int(key)] = str(key)
    return keys


KEYS_KEYS_CONSUMER = persistent_keys([ACTIONID.Key, ACTIONID.Consumer])
KEYS_KEYS_MOUSE_HSCROLL = persistent_keys([ACTIONID.Key, ACTIONID.Mouse, ACTIONID.Hscroll])

COLORS = _UnsortedNamedInts({
    # from Xorg rgb.txt,v 1.3 2000/08/17
    'red': 0xff0000,
    'orange': 0xffa500,
    'yellow': 0xffff00,
    'green': 0x00ff00,
    'blue': 0x0000ff,
    'purple': 0xa020f0,
    'violet': 0xee82ee,
    'black': 0x000000,
    'white': 0xffffff,
    'gray': 0xbebebe,
    'brown': 0xa52a2a,
    'cyan': 0x00ffff,
    'magenta': 0xff00ff,
    'pink': 0xffc0cb,
    'maroon': 0xb03060,
    'turquoise': 0x40e0d0,
    'gold': 0xffd700,
    'tan': 0xd2b48c,
    'snow': 0xfffafa,
    'ghost white': 0xf8f8ff,
    'white smoke': 0xf5f5f5,
    'gainsboro': 0xdcdcdc,
    'floral white': 0xfffaf0,
    'old lace': 0xfdf5e6,
    'linen': 0xfaf0e6,
    'antique white': 0xfaebd7,
    'papaya whip': 0xffefd5,
    'blanched almond': 0xffebcd,
    'bisque': 0xffe4c4,
    'peach puff': 0xffdab9,
    'navajo white': 0xffdead,
    'moccasin': 0xffe4b5,
    'cornsilk': 0xfff8dc,
    'ivory': 0xfffff0,
    'lemon chiffon': 0xfffacd,
    'seashell': 0xfff5ee,
    'honeydew': 0xf0fff0,
    'mint cream': 0xf5fffa,
    'azure': 0xf0ffff,
    'alice blue': 0xf0f8ff,
    'lavender': 0xe6e6fa,
    'lavender blush': 0xfff0f5,
    'misty rose': 0xffe4e1,
    'dark slate gray': 0x2f4f4f,
    'dim gray': 0x696969,
    'slate gray': 0x708090,
    'light slate gray': 0x778899,
    'light gray': 0xd3d3d3,
    'midnight blue': 0x191970,
    'navy blue': 0x000080,
    'cornflower blue': 0x6495ed,
    'dark slate blue': 0x483d8b,
    'slate blue': 0x6a5acd,
    'medium slate blue': 0x7b68ee,
    'light slate blue': 0x8470ff,
    'medium blue': 0x0000cd,
    'royal blue': 0x4169e1,
    'dodger blue': 0x1e90ff,
    'deep sky blue': 0x00bfff,
    'sky blue': 0x87ceeb,
    'light sky blue': 0x87cefa,
    'steel blue': 0x4682b4,
    'light steel blue': 0xb0c4de,
    'light blue': 0xadd8e6,
    'powder blue': 0xb0e0e6,
    'pale turquoise': 0xafeeee,
    'dark turquoise': 0x00ced1,
    'medium turquoise': 0x48d1cc,
    'light cyan': 0xe0ffff,
    'cadet blue': 0x5f9ea0,
    'medium aquamarine': 0x66cdaa,
    'aquamarine': 0x7fffd4,
    'dark green': 0x006400,
    'dark olive green': 0x556b2f,
    'dark sea green': 0x8fbc8f,
    'sea green': 0x2e8b57,
    'medium sea green': 0x3cb371,
    'light sea green': 0x20b2aa,
    'pale green': 0x98fb98,
    'spring green': 0x00ff7f,
    'lawn green': 0x7cfc00,
    'chartreuse': 0x7fff00,
    'medium spring green': 0x00fa9a,
    'green yellow': 0xadff2f,
    'lime green': 0x32cd32,
    'yellow green': 0x9acd32,
    'forest green': 0x228b22,
    'olive drab': 0x6b8e23,
    'dark khaki': 0xbdb76b,
    'khaki': 0xf0e68c,
    'pale goldenrod': 0xeee8aa,
    'light goldenrod yellow': 0xfafad2,
    'light yellow': 0xffffe0,
    'light goldenrod': 0xeedd82,
    'goldenrod': 0xdaa520,
    'dark goldenrod': 0xb8860b,
    'rosy brown': 0xbc8f8f,
    'indian red': 0xcd5c5c,
    'saddle brown': 0x8b4513,
    'sienna': 0xa0522d,
    'peru': 0xcd853f,
    'burlywood': 0xdeb887,
    'beige': 0xf5f5dc,
    'wheat': 0xf5deb3,
    'sandy brown': 0xf4a460,
    'chocolate': 0xd2691e,
    'firebrick': 0xb22222,
    'dark salmon': 0xe9967a,
    'salmon': 0xfa8072,
    'light salmon': 0xffa07a,
    'dark orange': 0xff8c00,
    'coral': 0xff7f50,
    'light coral': 0xf08080,
    'tomato': 0xff6347,
    'orange red': 0xff4500,
    'hot pink': 0xff69b4,
    'deep pink': 0xff1493,
    'light pink': 0xffb6c1,
    'pale violet red': 0xdb7093,
    'medium violet red': 0xc71585,
    'violet red': 0xd02090,
    'plum': 0xdda0dd,
    'orchid': 0xda70d6,
    'medium orchid': 0xba55d3,
    'dark orchid': 0x9932cc,
    'dark violet': 0x9400d3,
    'blue violet': 0x8a2be2,
    'medium purple': 0x9370db,
    'thistle': 0xd8bfd8,
    'dark gray': 0xa9a9a9,
    'dark blue': 0x00008b,
    'dark cyan': 0x008b8b,
    'dark magenta': 0x8b008b,
    'dark red': 0x8b0000,
    'light green': 0x90ee90,
})



# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
## Revisions Copyright (C) Contributors to the Solaar project.
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from gi.repository import Gtk
from solaar import NAME, __version__
from solaar.i18n import _

#
#
#

_dialog = None


def _create():
    about = Gtk.AboutDialog()

    about.set_program_name(NAME)
    about.set_version(__version__)
    about.set_comments(_('Manages Logitech receivers,\nkeyboards, mice, and tablets.'))
    about.set_logo_icon_name(NAME.lower())

    about.set_copyright('© 2012-2023 Daniel Pavel and contributors to the Solaar project')
    about.set_license_type(Gtk.License.GPL_2_0)

    about.set_authors(('Daniel Pavel http://github.com/pwr', ))
    try:
        about.add_credit_section(_('Additional Programming'), ('Filipe Laíns', 'Peter F. Patel-Schneider'))
        about.add_credit_section(_('GUI design'), ('Julien Gascard', 'Daniel Pavel'))
        about.add_credit_section(
            _('Testing'), (
                'Douglas Wagner',
                'Julien Gascard',
                'Peter Wu http://www.lekensteyn.nl/logitech-unifying.html',
            )
        )
        about.add_credit_section(
            _('Logitech documentation'), (
                'Julien Danjou http://julien.danjou.info/blog/2012/logitech-unifying-upower',
                'Nestor Lopez Casado http://drive.google.com/folderview?id=0BxbRzx7vEV7eWmgwazJ3NUFfQ28',
            )
        )
    except TypeError:
        # gtk3 < ~3.6.4 has incorrect gi bindings
        import logging
        logging.exception('failed to fully create the about dialog')
    except Exception:
        # the Gtk3 version may be too old, and the function does not exist
        import logging
        logging.exception('failed to fully create the about dialog')

    about.set_translator_credits(
        '\n'.join((
            'gogo (croatian)',
            'Papoteur, David Geiger, Damien Lallement (français)',
            'Michele Olivo (italiano)',
            'Adrian Piotrowicz (polski)',
            'Drovetto, JrBenito (Portuguese-BR)',
            'Daniel Pavel (română)',
            'Daniel Zippert, Emelie Snecker (svensk)',
            'Dimitriy Ryazantcev (Russian)',
            'El Jinete Sin Cabeza (Español)',
        ))
    )

    about.set_website('https://pwr-solaar.github.io/Solaar')
    about.set_website_label(NAME)

    about.connect('response', lambda x, y: x.hide())

    def _hide(dialog, event):
        dialog.hide()
        return True

    about.connect('delete-event', _hide)

    return about


def show_window(trigger=None):
    global _dialog
    if _dialog is None:
        _dialog = _create()
    _dialog.present()



# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Optional desktop notifications.

from solaar.i18n import _

#
#
#

try:
    import gi
    gi.require_version('Notify', '0.7')
    # this import is allowed to fail, in which case the entire feature is unavailable
    from gi.repository import GLib, Notify

    # assumed to be working since the import succeeded
    available = True

except (ValueError, ImportError):
    available = False

if available:
    import logging
    logger = logging.getLogger(__name__)

    from solaar import NAME

    from . import icons as _icons

    # cache references to shown notifications here, so if another status comes
    # while its notification is still visible we don't create another one
    _notifications = {}

    def init():
        """Init the notifications system."""
        global available
        if available:
            if not Notify.is_initted():
                if logger.isEnabledFor(logging.INFO):
                    logger.info('starting desktop notifications')
                try:
                    return Notify.init(NAME)
                except Exception:
                    logger.exception('initializing desktop notifications')
                    available = False
        return available and Notify.is_initted()

    def uninit():
        if available and Notify.is_initted():
            if logger.isEnabledFor(logging.INFO):
                logger.info('stopping desktop notifications')
            _notifications.clear()
            Notify.uninit()

    # def toggle(action):
    #     if action.get_active():
    #         init()
    #     else:
    #         uninit()
    #     action.set_sensitive(available)
    #     return action.get_active()

    def alert(reason, icon=None):
        assert reason

        if available and Notify.is_initted():
            n = _notifications.get(NAME)
            if n is None:
                n = _notifications[NAME] = Notify.Notification()

            # we need to use the filename here because the notifications daemon
            # is an external application that does not know about our icon sets
            icon_file = _icons.icon_file(NAME.lower()) if icon is None else _icons.icon_file(icon)

            n.update(NAME, reason, icon_file)
            n.set_urgency(Notify.Urgency.NORMAL)
            n.set_hint('desktop-entry', GLib.Variant('s', NAME.lower()))

            try:
                # if logger.isEnabledFor(logging.DEBUG):
                #     logger.debug("showing %s", n)
                n.show()
            except Exception:
                logger.exception('showing %s', n)

    def show(dev, reason=None, icon=None, progress=None):
        """Show a notification with title and text.
        Optionally displays the `progress` integer value
        in [0, 100] as a progress bar."""
        if available and Notify.is_initted():
            summary = dev.name

            # if a notification with same name is already visible, reuse it to avoid spamming
            n = _notifications.get(summary)
            if n is None:
                n = _notifications[summary] = Notify.Notification()

            if reason:
                message = reason
            elif dev.status is None:
                message = _('unpaired')
            elif bool(dev.status):
                message = dev.status.to_string() or _('connected')
            else:
                message = _('offline')

            # we need to use the filename here because the notifications daemon
            # is an external application that does not know about our icon sets
            icon_file = _icons.device_icon_file(dev.name, dev.kind) if icon is None else _icons.icon_file(icon)

            n.update(summary, message, icon_file)
            urgency = Notify.Urgency.LOW if dev.status else Notify.Urgency.NORMAL
            n.set_urgency(urgency)
            n.set_hint('desktop-entry', GLib.Variant('s', NAME.lower()))
            if progress:
                n.set_hint('value', GLib.Variant('i', progress))

            try:
                # if logger.isEnabledFor(logging.DEBUG):
                #     logger.debug("showing %s", n)
                n.show()
            except Exception:
                logger.exception('showing %s', n)

else:
    init = lambda: False
    uninit = lambda: None
    # toggle = lambda action: False
    alert = lambda reason: None
    show = lambda dev, reason=None: None



#!/usr/bin/env python3
# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging

from threading import Thread as _Thread

logger = logging.getLogger(__name__)

try:
    from Queue import Queue as _Queue
except ImportError:
    from queue import Queue as _Queue

#
#
#


class TaskRunner(_Thread):

    def __init__(self, name):
        super().__init__(name=name)
        self.daemon = True
        self.queue = _Queue(16)
        self.alive = False

    def __call__(self, function, *args, **kwargs):
        task = (function, args, kwargs)
        self.queue.put(task)

    def stop(self):
        self.alive = False
        self.queue.put(None)

    def run(self):
        self.alive = True

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('started')

        while self.alive:
            task = self.queue.get()
            if task:
                function, args, kwargs = task
                assert function
                try:
                    function(*args, **kwargs)
                except Exception:
                    logger.exception('calling %s', function)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('stopped')
